# Form implementation generated from reading ui file 'qt/aqt/forms/taglimit.ui'
#
# Created by: PyQt6 UI code generator 6.11.0
#
# WARNING: Any manual changes made to this file will be lost when pyuic6 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt6 import QtCore, QtGui, QtWidgets
from aqt.utils import tr



class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(361, 394)
        self.verticalLayout = QtWidgets.QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.activeCheck = QtWidgets.QCheckBox(parent=Dialog)
        self.activeCheck.setObjectName("activeCheck")
        self.verticalLayout.addWidget(self.activeCheck, 0, QtCore.Qt.AlignmentFlag.AlignLeft)
        self.activeList = QtWidgets.QListWidget(parent=Dialog)
        self.activeList.setEnabled(False)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(2)
        sizePolicy.setHeightForWidth(self.activeList.sizePolicy().hasHeightForWidth())
        self.activeList.setSizePolicy(sizePolicy)
        self.activeList.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.MultiSelection)
        self.activeList.setObjectName("activeList")
        self.verticalLayout.addWidget(self.activeList)
        self.label = QtWidgets.QLabel(parent=Dialog)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.inactiveList = QtWidgets.QListWidget(parent=Dialog)
        self.inactiveList.setEnabled(True)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Policy.Expanding, QtWidgets.QSizePolicy.Policy.Expanding)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(2)
        sizePolicy.setHeightForWidth(self.inactiveList.sizePolicy().hasHeightForWidth())
        self.inactiveList.setSizePolicy(sizePolicy)
        self.inactiveList.setSelectionMode(QtWidgets.QAbstractItemView.SelectionMode.MultiSelection)
        self.inactiveList.setObjectName("inactiveList")
        self.verticalLayout.addWidget(self.inactiveList)
        self.buttonBox = QtWidgets.QDialogButtonBox(parent=Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Orientation.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.StandardButton.Cancel|QtWidgets.QDialogButtonBox.StandardButton.Ok)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept) # type: ignore  # type: ignore
        self.buttonBox.rejected.connect(Dialog.reject) # type: ignore  # type: ignore
        self.activeCheck.toggled['bool'].connect(self.activeList.setEnabled) # type: ignore  # type: ignore
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(tr.custom_study_selective_study())
        self.activeCheck.setText(tr.custom_study_require_one_or_more_of_these())
        self.label.setText(tr.custom_study_select_tags_to_exclude())


# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

from collections.abc import Sequence

import aqt
import aqt.forms
from anki.cards import CardId
from anki.collection import (
    CARD_TYPE_NEW,
    Collection,
    Config,
    OpChanges,
    OpChangesWithCount,
    OpChangesWithId,
)
from anki.decks import DeckId
from anki.notes import NoteId
from anki.scheduler import CustomStudyRequest, FilteredDeckForUpdate, UnburyDeck
from anki.scheduler.base import ScheduleCardsAsNew
from anki.scheduler.v3 import CardAnswer
from anki.scheduler.v3 import Scheduler as V3Scheduler
from aqt.operations import CollectionOp
from aqt.qt import *
from aqt.utils import add_close_shortcut, disable_help_button, getText, tooltip, tr


def set_due_date_dialog(
    *,
    parent: QWidget,
    card_ids: Sequence[CardId],
    config_key: Config.String.V | None,
) -> CollectionOp[OpChanges] | None:
    assert aqt.mw
    if not card_ids:
        return None

    default_text = (
        aqt.mw.col.get_config_string(config_key) if config_key is not None else ""
    )
    prompt = "\n".join(
        [
            tr.scheduling_set_due_date_prompt(cards=len(card_ids)),
            tr.scheduling_set_due_date_prompt_hint(),
        ]
    )
    (days, success) = getText(
        prompt=prompt,
        parent=parent,
        default=default_text,
        title=tr.actions_set_due_date(),
    )
    if not success or not days.strip():
        return None
    else:
        return CollectionOp(
            parent, lambda col: col.sched.set_due_date(card_ids, days, config_key)
        ).success(
            lambda _: tooltip(
                tr.scheduling_set_due_date_done(cards=len(card_ids)),
                parent=parent,
            )
        )


def grade_now(
    *,
    parent: QWidget,
    card_ids: Sequence[CardId],
    ease: int,
    dialog: QDialog,
) -> None:
    if ease == 1:
        rating = CardAnswer.AGAIN
    elif ease == 2:
        rating = CardAnswer.HARD
    elif ease == 3:
        rating = CardAnswer.GOOD
    else:
        rating = CardAnswer.EASY
    CollectionOp(
        parent,
        lambda col: col._backend.grade_now(
            card_ids=card_ids,
            rating=rating,
        ),
    ).success(
        lambda _: tooltip(
            tr.scheduling_graded_cards_done(cards=len(card_ids)), parent=parent
        )
    ).run_in_background()
    dialog.accept()


def forget_cards(
    *,
    parent: QWidget,
    card_ids: Sequence[CardId],
    context: ScheduleCardsAsNew.Context.V | None = None,
) -> CollectionOp[OpChanges] | None:
    assert aqt.mw

    dialog = QDialog(parent)
    disable_help_button(dialog)
    add_close_shortcut(dialog)
    form = aqt.forms.forget.Ui_Dialog()
    form.setupUi(dialog)

    if context is not None:
        defaults = aqt.mw.col.sched.schedule_cards_as_new_defaults(context)
        form.restore_position.setChecked(defaults.restore_position)
        form.reset_counts.setChecked(defaults.reset_counts)

    if not dialog.exec():
        return None

    restore_position = form.restore_position.isChecked()
    reset_counts = form.reset_counts.isChecked()

    return CollectionOp(
        parent,
        lambda col: col.sched.schedule_cards_as_new(
            card_ids,
            restore_position=restore_position,
            reset_counts=reset_counts,
            context=context,
        ),
    ).success(
        lambda _: tooltip(
            tr.scheduling_forgot_cards(cards=len(card_ids)), parent=parent
        )
    )


def reposition_new_cards_dialog(
    *,
    parent: QWidget,
    card_ids: Sequence[CardId],
) -> CollectionOp[OpChangesWithCount] | None:
    from aqt import mw

    assert mw
    assert mw.col.db

    row = mw.col.db.first(
        f"select min(due), max(due) from cards where type={CARD_TYPE_NEW} and odid=0"
    )
    assert row
    (min_position, max_position) = row
    min_position = max(min_position or 0, 0)
    max_position = max_position or 0

    dialog = QDialog(parent)
    disable_help_button(dialog)
    add_close_shortcut(dialog)
    dialog.setWindowModality(Qt.WindowModality.WindowModal)
    form = aqt.forms.reposition.Ui_Dialog()
    form.setupUi(dialog)

    txt = tr.browsing_queue_top(val=min_position)
    txt += "\n" + tr.browsing_queue_bottom(val=max_position)
    form.label.setText(txt)

    form.start.selectAll()

    defaults = mw.col.sched.reposition_defaults()
    form.randomize.setChecked(defaults.random)
    form.shift.setChecked(defaults.shift)

    if not dialog.exec():
        return None

    start = form.start.value()
    step = form.step.value()
    randomize = form.randomize.isChecked()
    shift = form.shift.isChecked()

    return reposition_new_cards(
        parent=parent,
        card_ids=card_ids,
        starting_from=start,
        step_size=step,
        randomize=randomize,
        shift_existing=shift,
    )


def reposition_new_cards(
    *,
    parent: QWidget,
    card_ids: Sequence[CardId],
    starting_from: int,
    step_size: int,
    randomize: bool,
    shift_existing: bool,
) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(
        parent,
        lambda col: col.sched.reposition_new_cards(
            card_ids=card_ids,
            starting_from=starting_from,
            step_size=step_size,
            randomize=randomize,
            shift_existing=shift_existing,
        ),
    ).success(
        lambda out: tooltip(
            tr.browsing_changed_new_position(count=out.count), parent=parent
        )
    )


def suspend_cards(
    *,
    parent: QWidget,
    card_ids: Sequence[CardId],
) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(parent, lambda col: col.sched.suspend_cards(card_ids))


def suspend_note(
    *,
    parent: QWidget,
    note_ids: Sequence[NoteId],
) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(parent, lambda col: col.sched.suspend_notes(note_ids))


def unsuspend_cards(
    *, parent: QWidget, card_ids: Sequence[CardId]
) -> CollectionOp[OpChanges]:
    return CollectionOp(parent, lambda col: col.sched.unsuspend_cards(card_ids))


def bury_cards(
    *,
    parent: QWidget,
    card_ids: Sequence[CardId],
) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(parent, lambda col: col.sched.bury_cards(card_ids))


def bury_notes(
    *,
    parent: QWidget,
    note_ids: Sequence[NoteId],
) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(parent, lambda col: col.sched.bury_notes(note_ids))


def unbury_cards(
    *, parent: QWidget, card_ids: Sequence[CardId]
) -> CollectionOp[OpChanges]:
    return CollectionOp(parent, lambda col: col.sched.unbury_cards(card_ids))


def rebuild_filtered_deck(
    *, parent: QWidget, deck_id: DeckId
) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(parent, lambda col: col.sched.rebuild_filtered_deck(deck_id))


def empty_filtered_deck(*, parent: QWidget, deck_id: DeckId) -> CollectionOp[OpChanges]:
    return CollectionOp(parent, lambda col: col.sched.empty_filtered_deck(deck_id))


def add_or_update_filtered_deck(
    *,
    parent: QWidget,
    deck: FilteredDeckForUpdate,
) -> CollectionOp[OpChangesWithId]:
    return CollectionOp(parent, lambda col: col.sched.add_or_update_filtered_deck(deck))


def unbury_deck(
    *,
    parent: QWidget,
    deck_id: DeckId,
    mode: UnburyDeck.Mode.V = UnburyDeck.ALL,
) -> CollectionOp[OpChanges]:
    return CollectionOp(
        parent, lambda col: col.sched.unbury_deck(deck_id=deck_id, mode=mode)
    )


def answer_card(
    *,
    parent: QWidget,
    answer: CardAnswer,
) -> CollectionOp[OpChanges]:
    def answer_v3(col: Collection) -> OpChanges:
        assert isinstance(col.sched, V3Scheduler)
        return col.sched.answer_card(answer)

    return CollectionOp(parent, answer_v3)


def custom_study(
    *,
    parent: QWidget,
    request: CustomStudyRequest,
) -> CollectionOp[OpChanges]:
    return CollectionOp(parent, lambda col: col.sched.custom_study(request))



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

from collections.abc import Sequence

import aqt
import aqt.forms
import aqt.operations
from anki.notes import NoteId
from aqt import AnkiQt
from aqt.operations import QueryOp
from aqt.operations.note import find_and_replace
from aqt.operations.tag import find_and_replace_tag
from aqt.qt import *
from aqt.utils import (
    HelpPage,
    disable_help_button,
    openHelp,
    qconnect,
    restore_combo_history,
    restore_combo_index_for_session,
    restore_is_checked,
    restoreGeom,
    save_combo_history,
    save_combo_index_for_session,
    save_is_checked,
    saveGeom,
    tooltip,
    tr,
)


class FindAndReplaceDialog(QDialog):
    COMBO_NAME = "BrowserFindAndReplace"

    def __init__(
        self,
        parent: QWidget,
        *,
        mw: AnkiQt,
        note_ids: Sequence[NoteId],
        field: str | None = None,
    ) -> None:
        """
        If 'field' is passed, only this is added to the field selector.
        Otherwise, the fields belonging to the 'note_ids' are added.
        """
        super().__init__(parent)
        self.mw = mw
        self.note_ids = note_ids
        self.field_names: list[str] = []
        self._field = field

        if field:
            self._show([field])
        elif note_ids:
            # fetch field names and then show
            QueryOp(
                parent=mw,
                op=lambda col: col.field_names_for_note_ids(note_ids),
                success=self._show,
            ).run_in_background()
        else:
            self._show([])

    def _show(self, field_names: Sequence[str]) -> None:
        # add "all fields" and "tags" to the top of the list
        self.field_names = [
            tr.browsing_all_fields(),
            tr.editing_tags(),
        ] + list(field_names)

        disable_help_button(self)
        self.form = aqt.forms.findreplace.Ui_Dialog()
        self.form.setupUi(self)
        self.setWindowModality(Qt.WindowModality.WindowModal)

        self._find_history = restore_combo_history(
            self.form.find, self.COMBO_NAME + "Find"
        )

        find_completer = self.form.find.completer()
        assert find_completer is not None
        find_completer.setCaseSensitivity(Qt.CaseSensitivity.CaseSensitive)
        self._replace_history = restore_combo_history(
            self.form.replace, self.COMBO_NAME + "Replace"
        )
        replace_completer = self.form.replace.completer()
        assert replace_completer is not None
        replace_completer.setCaseSensitivity(Qt.CaseSensitivity.CaseSensitive)

        if not self.note_ids:
            # no selected notes to affect
            self.form.selected_notes.setChecked(False)
            self.form.selected_notes.setEnabled(False)
        elif self._field:
            self.form.selected_notes.setChecked(False)

        restore_is_checked(self.form.re, self.COMBO_NAME + "Regex")
        restore_is_checked(self.form.ignoreCase, self.COMBO_NAME + "ignoreCase")

        self.form.field.addItems(self.field_names)
        if self._field:
            self.form.field.setCurrentIndex(self.field_names.index(self._field))
        else:
            restore_combo_index_for_session(
                self.form.field, self.field_names, self.COMBO_NAME + "Field"
            )

        qconnect(self.form.buttonBox.helpRequested, self.show_help)

        restoreGeom(self, "findreplace")
        self.show()
        self.form.find.setFocus()

    def accept(self) -> None:
        saveGeom(self, "findreplace")
        save_combo_index_for_session(self.form.field, self.COMBO_NAME + "Field")

        search = save_combo_history(
            self.form.find, self._find_history, self.COMBO_NAME + "Find"
        )
        replace = save_combo_history(
            self.form.replace, self._replace_history, self.COMBO_NAME + "Replace"
        )
        regex = self.form.re.isChecked()
        match_case = not self.form.ignoreCase.isChecked()
        save_is_checked(self.form.re, self.COMBO_NAME + "Regex")
        save_is_checked(self.form.ignoreCase, self.COMBO_NAME + "ignoreCase")

        if not self.form.selected_notes.isChecked():
            # an empty list means *all* notes
            self.note_ids = []

        parent_widget = self.parentWidget()
        assert parent_widget is not None

        # tags?
        if self.form.field.currentIndex() == 1:
            op = find_and_replace_tag(
                parent=parent_widget,
                note_ids=self.note_ids,
                search=search,
                replacement=replace,
                regex=regex,
                match_case=match_case,
            )
        else:
            # fields
            if self.form.field.currentIndex() == 0:
                field = None
            else:
                field = self.field_names[self.form.field.currentIndex()]

            op = find_and_replace(
                parent=parent_widget,
                note_ids=self.note_ids,
                search=search,
                replacement=replace,
                regex=regex,
                field_name=field,
                match_case=match_case,
            )

        if not self.note_ids:
            op.success(
                lambda out: tooltip(
                    tr.browsing_notes_updated(count=out.count),
                    parent=self.parentWidget(),
                )
            )
        op.run_in_background()

        super().accept()

    def show_help(self) -> None:
        openHelp(HelpPage.BROWSING_FIND_AND_REPLACE)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html
# ruff: noqa: F401
from anki.utils import is_mac
from aqt.theme import theme_manager

from .item import SidebarItem, SidebarItemType
from .model import SidebarModel
from .searchbar import SidebarSearchBar
from .toolbar import SidebarTool, SidebarToolbar
from .tree import SidebarStage, SidebarTreeView



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html


def run_sync_server() -> None:
    import sys
    from os import environ as env

    from anki._backend import RustBackend

    env["RUST_LOG"] = env.get("RUST_LOG", "anki=info")

    try:
        RustBackend.syncserver()
    except Exception as exc:
        print("Sync server failed:", exc)
        sys.exit(1)
    sys.exit(0)


if __name__ == "__main__":
    run_sync_server()



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

# This file was automatically generated from _root-vars.scss
FONT_SIZE = {
    "comment": "",
    "light": "15px",
    "dark": "15px",
}
BORDER_RADIUS = {
    "comment": "Used to round corners of various UI elements",
    "light": "5px",
    "dark": "5px",
}
BORDER_RADIUS_MEDIUM = {
    "comment": "Used for container corners",
    "light": "12px",
    "dark": "12px",
}
BORDER_RADIUS_LARGE = {
    "comment": "Used for pill-shaped buttons",
    "light": "15px",
    "dark": "15px",
}
TRANSITION = {
    "comment": "Default duration of transitions in milliseconds",
    "light": "180",
    "dark": "180",
}
TRANSITION_MEDIUM = {
    "comment": "Slightly longer transition duration in milliseconds",
    "light": "500",
    "dark": "500",
}
TRANSITION_SLOW = {
    "comment": "Long transition duration in milliseconds",
    "light": "1000",
    "dark": "1000",
}
BLUR = {
    "comment": "Default background blur value",
    "light": "20px",
    "dark": "20px",
}



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

from collections.abc import Sequence

from anki.collection import OpChanges, OpChangesWithCount
from anki.decks import DeckId
from anki.notes import Note, NoteId
from aqt.operations import CollectionOp
from aqt.qt import QWidget
from aqt.utils import tooltip, tr


def add_note(
    *,
    parent: QWidget,
    note: Note,
    target_deck_id: DeckId,
) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(parent, lambda col: col.add_note(note, target_deck_id))


def update_note(*, parent: QWidget, note: Note) -> CollectionOp[OpChanges]:
    return CollectionOp(parent, lambda col: col.update_note(note))


def update_notes(*, parent: QWidget, notes: Sequence[Note]) -> CollectionOp[OpChanges]:
    return CollectionOp(parent, lambda col: col.update_notes(notes)).success(
        lambda _: tooltip(tr.browsing_cards_updated(count=len(notes)))
    )


def remove_notes(
    *,
    parent: QWidget,
    note_ids: Sequence[NoteId],
) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(parent, lambda col: col.remove_notes(note_ids)).success(
        lambda out: tooltip(tr.browsing_cards_deleted(count=out.count)),
    )


def find_and_replace(
    *,
    parent: QWidget,
    note_ids: Sequence[NoteId],
    search: str,
    replacement: str,
    regex: bool,
    field_name: str | None,
    match_case: bool,
) -> CollectionOp[OpChangesWithCount]:
    return CollectionOp(
        parent,
        lambda col: col.find_and_replace(
            note_ids=note_ids,
            search=search,
            replacement=replacement,
            regex=regex,
            field_name=field_name,
            match_case=match_case,
        ),
    ).success(
        lambda out: tooltip(
            tr.findreplace_notes_updated(changed=out.count, total=len(note_ids)),
            parent=parent,
        )
    )



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import functools
import json
import math
import re
from collections.abc import Callable, Sequence
from typing import Any, cast

from markdown import markdown

import aqt
import aqt.browser
import aqt.editor
import aqt.forms
import aqt.operations
from anki._legacy import deprecated
from anki.cards import Card, CardId
from anki.collection import Collection, Config, OpChanges, SearchNode
from anki.consts import *
from anki.decks import DeckId
from anki.errors import NotFoundError, SearchError
from anki.lang import without_unicode_isolation
from anki.models import NotetypeId
from anki.notes import NoteId
from anki.scheduler.base import ScheduleCardsAsNew
from anki.tags import MARKED_TAG
from anki.utils import is_mac
from aqt import AnkiQt, gui_hooks
from aqt.editor import Editor, EditorWebView
from aqt.errors import show_exception
from aqt.exporting import ExportDialog as LegacyExportDialog
from aqt.import_export.exporting import ExportDialog
from aqt.operations.card import set_card_deck, set_card_flag
from aqt.operations.collection import redo, undo
from aqt.operations.note import remove_notes
from aqt.operations.scheduling import (
    bury_cards,
    forget_cards,
    grade_now,
    reposition_new_cards_dialog,
    set_due_date_dialog,
    suspend_cards,
    unbury_cards,
    unsuspend_cards,
)
from aqt.operations.tag import (
    add_tags_to_notes,
    clear_unused_tags,
    remove_tags_from_notes,
)
from aqt.qt import *
from aqt.sound import av_player
from aqt.switch import Switch
from aqt.theme import WidgetStyle
from aqt.undo import UndoActionsInfo
from aqt.utils import (
    HelpPage,
    KeyboardModifiersPressed,
    add_close_shortcut,
    add_ellipsis_to_action_label,
    current_window,
    ensure_editor_saved,
    getTag,
    no_arg_trigger,
    openHelp,
    qtMenuShortcutWorkaround,
    restoreGeom,
    restoreSplitter,
    restoreState,
    saveGeom,
    saveSplitter,
    saveState,
    showWarning,
    skip_if_selection_is_empty,
    tooltip,
    tr,
)

from ..addcards import AddCards
from ..changenotetype import change_notetype_dialog
from .card_info import BrowserCardInfo
from .find_and_replace import FindAndReplaceDialog
from .layout import BrowserLayout, QSplitterHandleEventFilter
from .previewer import BrowserPreviewer as PreviewDialog
from .previewer import Previewer
from .sidebar import SidebarTreeView
from .table import Table


class MockModel:
    """This class only exists to support some legacy aliases."""

    def __init__(self, browser: aqt.browser.Browser) -> None:
        self.browser = browser

    @deprecated(replaced_by=aqt.operations.CollectionOp)
    def beginReset(self) -> None:
        self.browser.begin_reset()

    @deprecated(replaced_by=aqt.operations.CollectionOp)
    def endReset(self) -> None:
        self.browser.end_reset()

    @deprecated(replaced_by=aqt.operations.CollectionOp)
    def reset(self) -> None:
        self.browser.begin_reset()
        self.browser.end_reset()


class Browser(QMainWindow):
    mw: AnkiQt
    col: Collection
    editor: Editor | None
    table: Table

    def __init__(
        self,
        mw: AnkiQt,
        card: Card | None = None,
        search: tuple[str | SearchNode] | None = None,
    ) -> None:
        """
        card -- try to select the provided card after executing "search" or
                "deck:current" (if "search" was None)
        search -- set and perform search; caller must ensure validity
        """

        QMainWindow.__init__(self, None, Qt.WindowType.Window)
        self.mw = mw
        self.col = self.mw.col
        self.lastFilter = ""
        self.focusTo: int | None = None
        self._previewer: Previewer | None = None
        self._card_info = BrowserCardInfo(self.mw)
        self._closeEventHasCleanedUp = False
        self.auto_layout = True
        self.aspect_ratio = 0.0
        self.form = aqt.forms.browser.Ui_Dialog()
        self.form.setupUi(self)
        self.form.splitter.setChildrenCollapsible(False)
        splitter_handle_event_filter = QSplitterHandleEventFilter(self.form.splitter)

        splitter_handle = self.form.splitter.handle(1)
        assert splitter_handle is not None

        splitter_handle.installEventFilter(splitter_handle_event_filter)
        # set if exactly 1 row is selected; used by the previewer
        self.card: Card | None = None
        self.current_card: Card | None = None
        self.setupSidebar()
        self.setup_table()
        self.setupMenus()
        self.setupHooks()
        self.setupEditor()
        gui_hooks.browser_will_show(self)

        # restoreXXX() should be called after all child widgets have been created
        # and attached to QMainWindow
        self._editor_state_key = (
            "editorRTL"
            if self.layoutDirection() == Qt.LayoutDirection.RightToLeft
            else "editor"
        )
        restoreGeom(self, self._editor_state_key)
        restoreSplitter(self.form.splitter, "editor3")
        restoreState(self, self._editor_state_key)

        # responsive layout
        if self.height() != 0:
            self.aspect_ratio = self.width() / self.height()
        self.set_layout(self.mw.pm.browser_layout(), True)
        self.onSidebarVisibilityChange(not self.sidebarDockWidget.isHidden())
        # disable undo/redo
        self.on_undo_state_change(mw.undo_actions_info())
        # legacy alias
        self.model = MockModel(self)
        self.setupSearch(card, search)
        self.show()

    def on_operation_did_execute(
        self, changes: OpChanges, handler: object | None
    ) -> None:
        focused = current_window() == self
        self.table.op_executed(changes, handler, focused)
        self.sidebar.op_executed(changes, handler, focused)
        if changes.note_text:
            if handler is not self.editor:
                # fixme: this will leave the splitter shown, but with no current
                # note being edited
                assert self.editor is not None

                note = self.editor.note
                if note:
                    try:
                        note.load()
                    except NotFoundError:
                        self.editor.set_note(None)
                        return
                    self.editor.set_note(note)

        if changes.browser_table and changes.card:
            self.card = self.table.get_single_selected_card()
            self.current_card = self.table.get_current_card()
            self._update_card_info()
            self._update_current_actions()

        # changes.card is required for updating flag icon
        if changes.note_text or changes.card:
            self._renderPreview()

    def on_focus_change(self, new: QWidget | None, old: QWidget | None) -> None:
        if current_window() == self:
            self.setUpdatesEnabled(True)
            self.table.redraw_cells()
            self.sidebar.refresh_if_needed()

    def set_layout(self, mode: BrowserLayout, init: bool = False) -> None:
        self.mw.pm.set_browser_layout(mode)

        if mode == BrowserLayout.AUTO:
            self.auto_layout = True
            self.maybe_update_layout(self.aspect_ratio, True)
            self.form.actionLayoutAuto.setChecked(True)
            self.form.actionLayoutVertical.setChecked(False)
            self.form.actionLayoutHorizontal.setChecked(False)
            if not init:
                tooltip(tr.qt_misc_layout_auto_enabled())
        else:
            self.auto_layout = False
            self.form.actionLayoutAuto.setChecked(False)

            if mode == BrowserLayout.VERTICAL:
                self.form.splitter.setOrientation(Qt.Orientation.Vertical)
                self.form.actionLayoutVertical.setChecked(True)
                self.form.actionLayoutHorizontal.setChecked(False)
                if not init:
                    tooltip(tr.qt_misc_layout_vertical_enabled())

            elif mode == BrowserLayout.HORIZONTAL:
                self.form.splitter.setOrientation(Qt.Orientation.Horizontal)
                self.form.actionLayoutHorizontal.setChecked(True)
                self.form.actionLayoutVertical.setChecked(False)
                if not init:
                    tooltip(tr.qt_misc_layout_horizontal_enabled())

    def maybe_update_layout(self, aspect_ratio: float, force: bool = False) -> None:
        if force or math.floor(aspect_ratio) != math.floor(self.aspect_ratio):
            if aspect_ratio < 1:
                self.form.splitter.setOrientation(Qt.Orientation.Vertical)
            else:
                self.form.splitter.setOrientation(Qt.Orientation.Horizontal)

    def resizeEvent(self, event: QResizeEvent | None) -> None:
        assert event is not None

        if self.height() != 0:
            aspect_ratio = self.width() / self.height()

            if self.auto_layout:
                self.maybe_update_layout(aspect_ratio)

            self.aspect_ratio = aspect_ratio

        QMainWindow.resizeEvent(self, event)

    def get_active_note_type_id(self) -> NotetypeId | None:
        """
        If multiple cards are selected the note type will be derived
        from the final card selected
        """
        if current_note := self.table.get_current_note():
            return current_note.mid

        return None

    def add_card(self, deck_id: DeckId):
        add_cards = cast(AddCards, aqt.dialogs.open("AddCards", self.mw))
        add_cards.set_deck(deck_id)

        if note_type_id := self.get_active_note_type_id():
            add_cards.set_note_type(note_type_id)

    # If in the Browser we open Preview and press Ctrl+W there,
    # both Preview and Browser windows get closed by Qt out of the box.
    # We circumvent that behavior by only closing the currently active window
    def _handle_close(self):
        active_window = QApplication.activeWindow()
        if active_window and active_window != self:
            if isinstance(active_window, QDialog):
                active_window.reject()
            else:
                active_window.close()
        else:
            self.close()

    def setupMenus(self) -> None:
        # actions
        f = self.form

        # edit
        qconnect(f.actionUndo.triggered, self.undo)
        qconnect(f.actionRedo.triggered, self.redo)
        qconnect(f.actionInvertSelection.triggered, self.table.invert_selection)
        qconnect(f.actionSelectNotes.triggered, self.selectNotes)
        if not is_mac:
            f.actionClose.setVisible(False)
        qconnect(f.actionCreateFilteredDeck.triggered, self.createFilteredDeck)
        f.actionCreateFilteredDeck.setShortcuts(["Ctrl+G", "Ctrl+Alt+G"])

        # view
        qconnect(f.actionFullScreen.triggered, self.mw.on_toggle_full_screen)
        qconnect(
            f.actionZoomIn.triggered,
            lambda: self._editor_web_view().setZoomFactor(
                self._editor_web_view().zoomFactor() + 0.1
            ),
        )
        qconnect(
            f.actionZoomOut.triggered,
            lambda: self._editor_web_view().setZoomFactor(
                self._editor_web_view().zoomFactor() - 0.1
            ),
        )
        qconnect(
            f.actionResetZoom.triggered,
            lambda: self._editor_web_view().setZoomFactor(1),
        )
        qconnect(
            self.form.actionLayoutAuto.triggered,
            lambda: self.set_layout(BrowserLayout.AUTO),
        )
        qconnect(
            self.form.actionLayoutVertical.triggered,
            lambda: self.set_layout(BrowserLayout.VERTICAL),
        )
        qconnect(
            self.form.actionLayoutHorizontal.triggered,
            lambda: self.set_layout(BrowserLayout.HORIZONTAL),
        )

        # notes
        qconnect(f.actionAdd.triggered, self.mw.onAddCard)
        qconnect(f.actionCopy.triggered, self.on_create_copy)
        qconnect(f.actionAdd_Tags.triggered, self.add_tags_to_selected_notes)
        qconnect(f.actionRemove_Tags.triggered, self.remove_tags_from_selected_notes)
        qconnect(f.actionClear_Unused_Tags.triggered, self.clear_unused_tags)
        qconnect(f.actionToggle_Mark.triggered, self.toggle_mark_of_selected_notes)
        qconnect(f.actionChangeModel.triggered, self.onChangeModel)
        qconnect(f.actionFindDuplicates.triggered, self.onFindDupes)
        qconnect(f.actionFindReplace.triggered, self.onFindReplace)
        qconnect(f.actionManage_Note_Types.triggered, self.mw.onNoteTypes)
        qconnect(f.actionDelete.triggered, self.delete_selected_notes)

        # cards
        qconnect(f.actionChange_Deck.triggered, self.set_deck_of_selected_cards)
        qconnect(f.action_Info.triggered, self.showCardInfo)
        qconnect(f.actionReposition.triggered, self.reposition)
        qconnect(f.action_set_due_date.triggered, self.set_due_date)
        qconnect(f.action_grade_now.triggered, self.grade_now)
        qconnect(f.action_forget.triggered, self.forget_cards)
        qconnect(f.actionToggle_Suspend.triggered, self.suspend_selected_cards)
        qconnect(f.action_toggle_bury.triggered, self.bury_selected_cards)

        def set_flag_func(desired_flag: int) -> Callable:
            return lambda: self.set_flag_of_selected_cards(desired_flag)

        for flag in self.mw.flags.all():
            qconnect(
                getattr(self.form, flag.action).triggered, set_flag_func(flag.index)
            )
        self._update_flag_labels()
        qconnect(f.actionExport.triggered, self._on_export_notes)

        # jumps
        qconnect(f.actionPreviousCard.triggered, self.onPreviousCard)
        qconnect(f.actionNextCard.triggered, self.onNextCard)
        qconnect(f.actionFirstCard.triggered, self.onFirstCard)
        qconnect(f.actionLastCard.triggered, self.onLastCard)
        qconnect(f.actionFind.triggered, self.onFind)
        qconnect(f.actionNote.triggered, self.onNote)
        qconnect(f.actionSidebar.triggered, self.focusSidebar)
        qconnect(f.actionToggleSidebar.triggered, self.toggle_sidebar)
        qconnect(f.actionCardList.triggered, self.onCardList)

        # help
        qconnect(f.actionGuide.triggered, self.onHelp)

        # keyboard shortcut for shift+home/end
        self.pgUpCut = QShortcut(QKeySequence("Shift+Home"), self)
        qconnect(self.pgUpCut.activated, self.onFirstCard)
        self.pgDownCut = QShortcut(QKeySequence("Shift+End"), self)
        qconnect(self.pgDownCut.activated, self.onLastCard)

        # add-on hook
        gui_hooks.browser_menus_did_init(self)
        self.mw.maybeHideAccelerators(self)

        add_ellipsis_to_action_label(f.actionCopy)
        add_ellipsis_to_action_label(f.action_forget)
        add_ellipsis_to_action_label(f.action_grade_now)

    def _editor_web_view(self) -> EditorWebView:
        assert self.editor is not None
        editor_web_view = self.editor.web
        assert editor_web_view is not None
        return editor_web_view

    def closeEvent(self, evt: QCloseEvent | None) -> None:
        assert evt is not None

        if self._closeEventHasCleanedUp:
            evt.accept()
            return

        assert self.editor is not None

        self.editor.call_after_note_saved(self._closeWindow)
        evt.ignore()

    def _closeWindow(self) -> None:
        assert self.editor is not None

        self._cleanup_preview()
        self._card_info.close()
        self.editor.cleanup()
        self.table.cleanup()
        self.sidebar.cleanup()
        saveSplitter(self.form.splitter, "editor3")
        saveGeom(self, self._editor_state_key)
        saveState(self, self._editor_state_key)
        self.teardownHooks()
        self.mw.maybeReset()
        aqt.dialogs.markClosed("Browser")
        self._closeEventHasCleanedUp = True
        self.mw.deferred_delete_and_garbage_collect(self)
        self.close()

    @ensure_editor_saved
    def closeWithCallback(self, onsuccess: Callable) -> None:
        self._closeWindow()
        onsuccess()

    def keyPressEvent(self, evt: QKeyEvent | None) -> None:
        assert evt is not None

        if evt.key() == Qt.Key.Key_Escape:
            self.close()
        else:
            super().keyPressEvent(evt)

    def reopen(
        self,
        _mw: AnkiQt,
        card: Card | None = None,
        search: tuple[str | SearchNode] | None = None,
    ) -> None:
        if search is not None:
            self.search_for_terms(*search)
            self.form.searchEdit.setFocus()
        if card is not None:
            if search is None:
                # implicitly assume 'card' is in the current deck
                self._default_search(card)
                self.form.searchEdit.setFocus()
            self.table.select_single_card(card.id)

    # Searching
    ######################################################################

    def setupSearch(
        self,
        card: Card | None = None,
        search: tuple[str | SearchNode] | None = None,
    ) -> None:
        assert self.mw.pm.profile is not None

        line_edit = self._line_edit()
        qconnect(line_edit.returnPressed, self.onSearchActivated)
        self.form.searchEdit.setCompleter(None)
        line_edit.setPlaceholderText(tr.browsing_search_bar_hint())
        line_edit.setMaxLength(2000000)
        self.form.searchEdit.addItems(
            [""] + self.mw.pm.profile.get("searchHistory", [])
        )
        if search is not None:
            self.search_for_terms(*search)
        else:
            self._default_search(card)
        self.form.searchEdit.setFocus()
        if card:
            self.table.select_single_card(card.id)

    # search triggered by user
    @ensure_editor_saved
    def onSearchActivated(self) -> None:
        text = self.current_search()
        try:
            normed = self.col.build_search_string(text)
        except SearchError as err:
            showWarning(markdown(str(err)))
        except Exception as err:
            showWarning(str(err))
        else:
            self.search_for(normed)
            self.update_history()

    def search_for(self, search: str, prompt: str | None = None) -> None:
        """Keep track of search string so that we reuse identical search when
        refreshing, rather than whatever is currently in the search field.
        Optionally set the search bar to a different text than the actual search.
        """

        self._lastSearchTxt = search
        prompt = search if prompt is None else prompt
        self.form.searchEdit.setCurrentIndex(-1)
        self._line_edit().setText(prompt)
        self.search()

    def current_search(self) -> str:
        return self._line_edit().text()

    def search(self) -> None:
        """Search triggered programmatically. Caller must have saved note first."""

        try:
            self.table.search(self._lastSearchTxt)
        except Exception as err:
            showWarning(str(err))

    def update_history(self) -> None:
        assert self.mw.pm.profile is not None

        sh = self.mw.pm.profile.get("searchHistory", [])
        if self._lastSearchTxt in sh:
            sh.remove(self._lastSearchTxt)
        sh.insert(0, self._lastSearchTxt)
        sh = sh[:30]
        self.form.searchEdit.clear()
        self.form.searchEdit.addItems(sh)
        self.mw.pm.profile["searchHistory"] = sh

    def updateTitle(self) -> None:
        selected = self.table.len_selection()
        cur = self.table.len()
        tr_title = (
            tr.browsing_window_title_notes
            if self.table.is_notes_mode()
            else tr.browsing_window_title
        )
        self.setWindowTitle(
            without_unicode_isolation(tr_title(total=cur, selected=selected))
        )

    def search_for_terms(self, *search_terms: str | SearchNode) -> None:
        search = self.col.build_search_string(*search_terms)
        self.form.searchEdit.setEditText(search)
        self.onSearchActivated()

    def _default_search(self, card: Card | None = None) -> None:
        default = self.col.get_config_string(Config.String.DEFAULT_SEARCH_TEXT)
        if default.strip():
            search = default
            prompt = default
        else:
            search = self.col.build_search_string(SearchNode(deck="current"))
            prompt = ""
        if card is not None:
            search = gui_hooks.default_search(search, card)
        self.search_for(search, prompt)

    def onReset(self) -> None:
        self.sidebar.refresh()
        self.begin_reset()
        self.end_reset()

    # caller must have called editor.saveNow() before calling this or .reset()
    def begin_reset(self) -> None:
        assert self.editor is not None

        self.editor.set_note(None, hide=False)
        self.mw.progress.start()
        self.table.begin_reset()

    def end_reset(self) -> None:
        self.table.end_reset()
        self.mw.progress.finish()

    # Table & Editor
    ######################################################################

    def setup_table(self) -> None:
        self.table = Table(self)
        self.table.set_view(self.form.tableView)
        self._switch = switch = Switch(12, tr.browsing_cards(), tr.browsing_notes())
        switch.setChecked(self.table.is_notes_mode())
        switch.setToolTip(tr.browsing_toggle_showing_cards_notes())
        qconnect(self.form.action_toggle_mode.triggered, switch.toggle)
        qconnect(switch.toggled, self.on_table_state_changed)
        self.form.gridLayout.addWidget(switch, 0, 0)

    def setupEditor(self) -> None:
        QShortcut(QKeySequence("Ctrl+Shift+P"), self, self.onTogglePreview)

        def add_preview_button(editor: Editor) -> None:
            editor._links["preview"] = lambda _editor: self.onTogglePreview()

        gui_hooks.editor_did_init.append(add_preview_button)
        self.editor = aqt.editor.Editor(
            self.mw,
            self.form.fieldsArea,
            self,
            editor_mode=aqt.editor.EditorMode.BROWSER,
        )
        gui_hooks.editor_did_init.remove(add_preview_button)

    @ensure_editor_saved
    def on_all_or_selected_rows_changed(self) -> None:
        """Called after the selected or all rows (searching, toggling mode) have
        changed. Update window title, card preview, context actions, and editor.
        """
        if self._closeEventHasCleanedUp:
            return

        self.updateTitle()
        # if there is only one selected card, use it in the editor
        # it might differ from the current card
        self.card = self.table.get_single_selected_card()
        self.singleCard = bool(self.card)

        splitter_widget = self.form.splitter.widget(1)
        assert splitter_widget is not None

        splitter_widget.setVisible(self.singleCard)

        assert self.editor is not None

        if self.singleCard:
            assert self.card is not None

            self.editor.set_note(self.card.note(), focusTo=self.focusTo)
            self.focusTo = None
            self.editor.card = self.card
        else:
            self.editor.set_note(None)
        self._renderPreview()
        self._update_row_actions()
        self._update_selection_actions()
        gui_hooks.browser_did_change_row(self)

    @deprecated(info="please use on_all_or_selected_rows_changed() instead.")
    def onRowChanged(self, *args: Any) -> None:
        self.on_all_or_selected_rows_changed()

    def on_current_row_changed(self) -> None:
        """Called after the row of the current element has changed."""
        if self._closeEventHasCleanedUp:
            return
        self.current_card = self.table.get_current_card()
        self._update_current_actions()
        self._update_card_info()

    def _update_row_actions(self) -> None:
        has_rows = bool(self.table.len())
        self.form.actionSelectAll.setEnabled(has_rows)
        self.form.actionInvertSelection.setEnabled(has_rows)
        self.form.actionFirstCard.setEnabled(has_rows)
        self.form.actionLastCard.setEnabled(has_rows)

    def _update_selection_actions(self) -> None:
        has_selection = bool(self.table.len_selection())
        self.form.actionSelectNotes.setEnabled(has_selection)
        self.form.actionExport.setEnabled(has_selection)
        self.form.actionAdd_Tags.setEnabled(has_selection)
        self.form.actionRemove_Tags.setEnabled(has_selection)
        self.form.actionToggle_Mark.setEnabled(has_selection)
        self.form.actionChangeModel.setEnabled(has_selection)
        self.form.actionDelete.setEnabled(has_selection)
        self.form.actionChange_Deck.setEnabled(has_selection)
        self.form.action_set_due_date.setEnabled(has_selection)
        self.form.action_forget.setEnabled(has_selection)
        self.form.actionReposition.setEnabled(has_selection)
        self.form.actionToggle_Suspend.setEnabled(has_selection)
        self.form.action_toggle_bury.setEnabled(has_selection)
        self.form.menuFlag.setEnabled(has_selection)

    def _update_current_actions(self) -> None:
        self._update_flags_menu()
        self._update_toggle_bury_action()
        self._update_toggle_mark_action()
        self._update_toggle_suspend_action()
        self.form.actionCopy.setEnabled(self.table.has_current())
        self.form.action_Info.setEnabled(self.table.has_current())
        self.form.actionPreviousCard.setEnabled(self.table.has_previous())
        self.form.actionNextCard.setEnabled(self.table.has_next())

    @ensure_editor_saved
    def on_table_state_changed(self, checked: bool) -> None:
        self.mw.progress.start()
        try:
            self.table.toggle_state(checked, self._lastSearchTxt)
        except Exception as err:
            self.mw.progress.finish()
            self._switch.blockSignals(True)
            self._switch.toggle()
            self._switch.blockSignals(False)
            show_exception(parent=self, exception=err)
        else:
            self.mw.progress.finish()

    # Sidebar
    ######################################################################

    def setupSidebar(self) -> None:
        dw = self.sidebarDockWidget = QDockWidget(tr.browsing_sidebar(), self)
        dw.setFeatures(QDockWidget.DockWidgetFeature.DockWidgetClosable)
        dw.setObjectName("Sidebar")
        dock_area = (
            Qt.DockWidgetArea.RightDockWidgetArea
            if self.layoutDirection() == Qt.LayoutDirection.RightToLeft
            else Qt.DockWidgetArea.LeftDockWidgetArea
        )
        dw.setAllowedAreas(dock_area)

        self.sidebar = SidebarTreeView(self)
        self.sidebarTree = self.sidebar  # legacy alias
        dw.setWidget(self.sidebar)
        qconnect(
            self.form.actionSidebarFilter.triggered,
            self.focusSidebarSearchBar,
        )
        qconnect(dw.visibilityChanged, self.onSidebarVisibilityChange)
        grid = QGridLayout()
        grid.addWidget(self.sidebar.searchBar, 0, 0)
        grid.addWidget(self.sidebar.toolbar, 0, 1)
        grid.addWidget(self.sidebar, 1, 0, 1, 2)
        grid.setContentsMargins(8, 4, 0, 0)
        grid.setSpacing(0)
        w = QWidget()
        w.setLayout(grid)
        dw.setWidget(w)
        self.sidebarDockWidget.setFloating(False)

        self.sidebarDockWidget.setTitleBarWidget(QWidget())
        self.addDockWidget(dock_area, dw)

        # schedule sidebar to refresh after browser window has loaded, so the
        # UI is more responsive
        self.mw.progress.timer(10, self.sidebar.refresh, False, parent=self.sidebar)

    def showSidebar(self, show: bool = True) -> None:
        self.sidebarDockWidget.setVisible(show)

    def onSidebarVisibilityChange(self, visible):
        margins = self.form.verticalLayout_3.contentsMargins()
        skip_left_margin = visible and not (
            is_mac and aqt.mw.pm.get_widget_style() == WidgetStyle.NATIVE
        )
        margins.setLeft(0 if skip_left_margin else margins.right())
        self.form.verticalLayout_3.setContentsMargins(margins)

        if visible:
            self.sidebar.refresh()

    def focusSidebar(self) -> None:
        self.showSidebar()
        self.sidebar.setFocus()

    def focusSidebarSearchBar(self) -> None:
        self.showSidebar()
        self.sidebar.searchBar.setFocus()

    def toggle_sidebar(self) -> None:
        self.showSidebar(not self.sidebarDockWidget.isVisible())

    # legacy

    def setFilter(self, *terms: str) -> None:
        self.sidebar.update_search(*terms)

    # Info
    ######################################################################

    def showCardInfo(self) -> None:
        self._card_info.show()

    def _update_card_info(self) -> None:
        self._card_info.set_card(self.current_card)

    # Menu helpers
    ######################################################################

    def selected_cards(self) -> Sequence[CardId]:
        return self.table.get_selected_card_ids()

    def selected_notes(self) -> Sequence[NoteId]:
        return self.table.get_selected_note_ids()

    def selectedNotesAsCards(self) -> Sequence[CardId]:
        return self.table.get_card_ids_from_selected_note_ids()

    def onHelp(self) -> None:
        openHelp(HelpPage.BROWSING)

    # legacy

    selectedCards = selected_cards
    selectedNotes = selected_notes

    # Misc menu options
    ######################################################################

    def on_create_copy(self) -> None:
        if note := self.table.get_current_note():
            current_card = self.table.get_current_card()
            assert current_card is not None

            deck_id = current_card.current_deck_id()
            aqt.dialogs.open("AddCards", self.mw).set_note(note, deck_id)

    @no_arg_trigger
    @skip_if_selection_is_empty
    @ensure_editor_saved
    def onChangeModel(self) -> None:
        ids = self.selected_notes()
        change_notetype_dialog(parent=self, note_ids=ids)

    def createFilteredDeck(self) -> None:
        search = self.current_search()
        if KeyboardModifiersPressed().alt:
            aqt.dialogs.open("FilteredDeckConfigDialog", self.mw, search_2=search)
        else:
            aqt.dialogs.open("FilteredDeckConfigDialog", self.mw, search=search)

    # Preview
    ######################################################################

    def onTogglePreview(self) -> None:
        assert self.editor is not None

        if self._previewer:
            self._previewer.close()
        elif self.editor.note:
            self._previewer = PreviewDialog(self, self.mw, self._on_preview_closed)
            self._previewer.open()
            self.toggle_preview_button_state(True)

    def _renderPreview(self) -> None:
        if self._previewer:
            if self.singleCard:
                self._previewer.render_card()
            else:
                self.onTogglePreview()

    def toggle_preview_button_state(self, active: bool) -> None:
        assert self.editor is not None

        if self.editor.web:
            self.editor.web.eval(f"togglePreviewButtonState({json.dumps(active)});")

    def _cleanup_preview(self) -> None:
        if self._previewer:
            self._previewer.cancel_timer()
            self._previewer.close()

    def _on_preview_closed(self) -> None:
        av_player.stop_and_clear_queue()
        self.toggle_preview_button_state(False)
        self._previewer = None

    # Card deletion
    ######################################################################

    @no_arg_trigger
    @skip_if_selection_is_empty
    def delete_selected_notes(self) -> None:
        # ensure deletion is not accidentally triggered when the user is focused
        # in the editing screen or search bar
        focus = self.focusWidget()
        if focus != self.form.tableView:
            return

        assert self.editor is not None

        self.editor.set_note(None)
        nids = self.table.to_row_of_unselected_note()
        remove_notes(parent=self, note_ids=nids).run_in_background()

    # legacy

    deleteNotes = delete_selected_notes

    # Deck change
    ######################################################################

    @no_arg_trigger
    @skip_if_selection_is_empty
    @ensure_editor_saved
    def set_deck_of_selected_cards(self) -> None:
        from aqt.studydeck import StudyDeck

        assert self.mw.col is not None
        assert self.mw.col.db is not None

        cids = self.table.get_selected_card_ids()
        did = self.mw.col.db.scalar("select did from cards where id = ?", cids[0])

        deck_dict = self.mw.col.decks.get(did)
        assert deck_dict is not None

        current = deck_dict["name"]

        def callback(ret: StudyDeck) -> None:
            if not ret.name:
                return
            did = self.col.decks.id(ret.name)

            assert did is not None

            set_card_deck(parent=self, card_ids=cids, deck_id=did).run_in_background()

        StudyDeck(
            self.mw,
            current=current,
            accept=tr.browsing_move_cards(),
            title=tr.browsing_change_deck(),
            help=HelpPage.BROWSING,
            parent=self,
            callback=callback,
        )

    # legacy

    setDeck = set_deck_of_selected_cards

    # Tags
    ######################################################################

    @no_arg_trigger
    @skip_if_selection_is_empty
    @ensure_editor_saved
    def add_tags_to_selected_notes(
        self,
        tags: str | None = None,
    ) -> None:
        "Shows prompt if tags not provided."
        if not (tags := tags or self._prompt_for_tags(tr.browsing_enter_tags_to_add())):
            return

        space_separated_tags = re.sub(r"[ \n\t\v]+", " ", tags)
        add_tags_to_notes(
            parent=self,
            note_ids=self.selected_notes(),
            space_separated_tags=space_separated_tags,
        ).run_in_background(initiator=self)

    @no_arg_trigger
    @skip_if_selection_is_empty
    @ensure_editor_saved
    def remove_tags_from_selected_notes(self, tags: str | None = None) -> None:
        "Shows prompt if tags not provided."
        if not (
            tags := tags or self._prompt_for_tags(tr.browsing_enter_tags_to_delete())
        ):
            return

        remove_tags_from_notes(
            parent=self, note_ids=self.selected_notes(), space_separated_tags=tags
        ).run_in_background(initiator=self)

    def _prompt_for_tags(self, prompt: str) -> str | None:
        (tags, ok) = getTag(self, self.col, prompt)
        if not ok:
            return None
        else:
            return tags

    @no_arg_trigger
    @ensure_editor_saved
    def clear_unused_tags(self) -> None:
        clear_unused_tags(parent=self).run_in_background()

    addTags = add_tags_to_selected_notes
    deleteTags = remove_tags_from_selected_notes
    clearUnusedTags = clear_unused_tags

    # Suspending
    ######################################################################

    def _update_toggle_suspend_action(self) -> None:
        is_suspended = bool(
            self.current_card and self.current_card.queue == QUEUE_TYPE_SUSPENDED
        )
        self.form.actionToggle_Suspend.setChecked(is_suspended)

    @skip_if_selection_is_empty
    @ensure_editor_saved
    def suspend_selected_cards(self, checked: bool) -> None:
        cids = self.selected_cards()
        if checked:
            suspend_cards(parent=self, card_ids=cids).run_in_background()
        else:
            unsuspend_cards(parent=self.mw, card_ids=cids).run_in_background()

    # Burying
    ######################################################################

    def _update_toggle_bury_action(self) -> None:
        is_buried = bool(
            self.current_card
            and self.current_card.queue
            in (QUEUE_TYPE_MANUALLY_BURIED, QUEUE_TYPE_SIBLING_BURIED)
        )
        self.form.action_toggle_bury.setChecked(is_buried)

    @skip_if_selection_is_empty
    @ensure_editor_saved
    def bury_selected_cards(self, checked: bool) -> None:
        cids = self.selected_cards()
        if checked:
            bury_cards(parent=self, card_ids=cids).run_in_background()
        else:
            unbury_cards(parent=self.mw, card_ids=cids).run_in_background()

    # Exporting
    ######################################################################

    @no_arg_trigger
    @skip_if_selection_is_empty
    def _on_export_notes(self) -> None:
        if not self.mw.pm.legacy_import_export():
            nids = self.selected_notes()
            ExportDialog(self.mw, nids=nids, parent=self)
        else:
            cids = self.selectedNotesAsCards()
            LegacyExportDialog(self.mw, cids=list(cids), parent=self)

    # Flags & Marking
    ######################################################################

    @skip_if_selection_is_empty
    @ensure_editor_saved
    def set_flag_of_selected_cards(self, flag: int) -> None:
        if not self.current_card:
            return

        # flag needs toggling off?
        if flag == self.current_card.user_flag():
            flag = 0

        set_card_flag(
            parent=self, card_ids=self.selected_cards(), flag=flag
        ).run_in_background()

    def _update_flags_menu(self) -> None:
        flag = self.current_card and self.current_card.user_flag()
        flag = flag or 0

        for f in self.mw.flags.all():
            getattr(self.form, f.action).setChecked(flag == f.index)

        qtMenuShortcutWorkaround(self.form.menuFlag)

    def _update_flag_labels(self) -> None:
        for flag in self.mw.flags.all():
            getattr(self.form, flag.action).setText(flag.label)

    def toggle_mark_of_selected_notes(self, checked: bool) -> None:
        if checked:
            self.add_tags_to_selected_notes(tags=MARKED_TAG)
        else:
            self.remove_tags_from_selected_notes(tags=MARKED_TAG)

    def _update_toggle_mark_action(self) -> None:
        is_marked = bool(
            self.current_card and self.current_card.note().has_tag(MARKED_TAG)
        )
        self.form.actionToggle_Mark.setChecked(is_marked)

    # Scheduling
    ######################################################################

    @no_arg_trigger
    @skip_if_selection_is_empty
    @ensure_editor_saved
    def reposition(self) -> None:
        if op := reposition_new_cards_dialog(
            parent=self, card_ids=self.selected_cards()
        ):
            op.run_in_background()

    @no_arg_trigger
    @skip_if_selection_is_empty
    @ensure_editor_saved
    def set_due_date(self) -> None:
        if op := set_due_date_dialog(
            parent=self,
            card_ids=self.selected_cards(),
            config_key=Config.String.SET_DUE_BROWSER,
        ):
            op.run_in_background()

    @no_arg_trigger
    @skip_if_selection_is_empty
    @ensure_editor_saved
    def forget_cards(self) -> None:
        if op := forget_cards(
            parent=self,
            card_ids=self.selected_cards(),
            context=ScheduleCardsAsNew.Context.BROWSER,
        ):
            op.run_in_background()

    @no_arg_trigger
    @skip_if_selection_is_empty
    @ensure_editor_saved
    def grade_now(self) -> None:
        """Show dialog to grade selected cards."""
        dialog = QDialog(self)
        dialog.setWindowTitle(tr.actions_grade_now())
        layout = QHBoxLayout()
        dialog.setLayout(layout)
        add_close_shortcut(dialog)

        # Add grade buttons
        for ease, label in [
            (1, tr.studying_again()),
            (2, tr.studying_hard()),
            (3, tr.studying_good()),
            (4, tr.studying_easy()),
        ]:
            btn = QPushButton(label)
            qconnect(
                btn.clicked,
                functools.partial(
                    grade_now,
                    parent=self,
                    card_ids=self.selected_cards(),
                    ease=ease,
                    dialog=dialog,
                ),
            )
            if key := aqt.mw.pm.get_answer_key(ease):
                QShortcut(key, dialog, activated=btn.click)  # type: ignore
                btn.setToolTip(tr.actions_shortcut_key(key))
            layout.addWidget(btn)

        # Add cancel button
        cancel_btn = QPushButton(tr.actions_cancel())
        qconnect(cancel_btn.clicked, dialog.reject)
        layout.addWidget(cancel_btn)

        dialog.exec()

    # Edit: selection
    ######################################################################

    @no_arg_trigger
    @skip_if_selection_is_empty
    @ensure_editor_saved
    def selectNotes(self) -> None:
        nids = self.selected_notes()
        # clear the selection so we don't waste energy preserving it
        self.table.clear_selection()
        search = self.col.build_search_string(
            SearchNode(nids=SearchNode.IdList(ids=nids))
        )
        self.search_for(search)
        self.table.select_all()

    # Hooks
    ######################################################################

    def setupHooks(self) -> None:
        gui_hooks.undo_state_did_change.append(self.on_undo_state_change)
        gui_hooks.backend_will_block.append(self.table.on_backend_will_block)
        gui_hooks.backend_did_block.append(self.table.on_backend_did_block)
        gui_hooks.operation_did_execute.append(self.on_operation_did_execute)
        gui_hooks.focus_did_change.append(self.on_focus_change)
        gui_hooks.flag_label_did_change.append(self._update_flag_labels)
        gui_hooks.collection_will_temporarily_close.append(self._on_temporary_close)

    def teardownHooks(self) -> None:
        gui_hooks.undo_state_did_change.remove(self.on_undo_state_change)
        gui_hooks.backend_will_block.remove(self.table.on_backend_will_block)
        gui_hooks.backend_did_block.remove(self.table.on_backend_did_block)
        gui_hooks.operation_did_execute.remove(self.on_operation_did_execute)
        gui_hooks.focus_did_change.remove(self.on_focus_change)
        gui_hooks.flag_label_did_change.remove(self._update_flag_labels)
        gui_hooks.collection_will_temporarily_close.remove(self._on_temporary_close)

    def _on_temporary_close(self, col: Collection) -> None:
        # we could reload browser columns in the future; for now we just close
        self.close()

    # Undo
    ######################################################################

    def undo(self) -> None:
        undo(parent=self)

    def redo(self) -> None:
        redo(parent=self)

    def on_undo_state_change(self, info: UndoActionsInfo) -> None:
        self.form.actionUndo.setText(info.undo_text)
        self.form.actionUndo.setEnabled(info.can_undo)
        self.form.actionRedo.setText(info.redo_text)
        self.form.actionRedo.setEnabled(info.can_redo)
        self.form.actionRedo.setVisible(info.show_redo)

    # Edit: replacing
    ######################################################################

    @no_arg_trigger
    @ensure_editor_saved
    def onFindReplace(self) -> None:
        FindAndReplaceDialog(self, mw=self.mw, note_ids=self.selected_notes())

    # Edit: finding dupes
    ######################################################################

    @no_arg_trigger
    @ensure_editor_saved
    def onFindDupes(self) -> None:
        from aqt.browser.find_duplicates import FindDuplicatesDialog

        FindDuplicatesDialog(browser=self, mw=self.mw)

    # Jumping
    ######################################################################

    def has_previous_card(self) -> bool:
        return self.table.has_previous()

    def has_next_card(self) -> bool:
        return self.table.has_next()

    def onPreviousCard(self) -> None:
        assert self.editor is not None

        self.focusTo = self.editor.currentField
        self.editor.call_after_note_saved(self.table.to_previous_row)

    def onNextCard(self) -> None:
        assert self.editor is not None

        self.focusTo = self.editor.currentField
        self.editor.call_after_note_saved(self.table.to_next_row)

    def onFirstCard(self) -> None:
        self.table.to_first_row()

    def onLastCard(self) -> None:
        self.table.to_last_row()

    def onFind(self) -> None:
        self.form.searchEdit.setFocus()
        self._line_edit().selectAll()

    def onNote(self) -> None:
        def cb():
            assert self.editor is not None and self.editor.web is not None
            self.editor.web.setFocus()
            self.editor.loadNote(focusTo=0)

        assert self.editor is not None
        self.editor.call_after_note_saved(cb)

    def onCardList(self) -> None:
        self.form.tableView.setFocus()

    def _line_edit(self) -> QLineEdit:
        line_edit = self.form.searchEdit.lineEdit()
        assert line_edit is not None
        return line_edit



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/scheduler.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/scheduler.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from anki import generic_pb2 as anki_dot_generic__pb2
from anki import cards_pb2 as anki_dot_cards__pb2
from anki import decks_pb2 as anki_dot_decks__pb2
from anki import collection_pb2 as anki_dot_collection__pb2
from anki import config_pb2 as anki_dot_config__pb2
from anki import deck_config_pb2 as anki_dot_deck__config__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14\x61nki/scheduler.proto\x12\x0e\x61nki.scheduler\x1a\x12\x61nki/generic.proto\x1a\x10\x61nki/cards.proto\x1a\x10\x61nki/decks.proto\x1a\x15\x61nki/collection.proto\x1a\x11\x61nki/config.proto\x1a\x16\x61nki/deck_config.proto\"\xd5\t\n\x0fSchedulingState\x12\x38\n\x06normal\x18\x01 \x01(\x0b\x32&.anki.scheduler.SchedulingState.NormalH\x00\x12<\n\x08\x66iltered\x18\x02 \x01(\x0b\x32(.anki.scheduler.SchedulingState.FilteredH\x00\x12\x18\n\x0b\x63ustom_data\x18\x03 \x01(\tH\x01\x88\x01\x01\x1a\x17\n\x03New\x12\x10\n\x08position\x18\x01 \x01(\r\x1a\x9a\x01\n\x08Learning\x12\x17\n\x0fremaining_steps\x18\x01 \x01(\r\x12\x16\n\x0escheduled_secs\x18\x02 \x01(\r\x12\x14\n\x0c\x65lapsed_secs\x18\x03 \x01(\r\x12\x36\n\x0cmemory_state\x18\x06 \x01(\x0b\x32\x1b.anki.cards.FsrsMemoryStateH\x00\x88\x01\x01\x42\x0f\n\r_memory_state\x1a\xb5\x01\n\x06Review\x12\x16\n\x0escheduled_days\x18\x01 \x01(\r\x12\x14\n\x0c\x65lapsed_days\x18\x02 \x01(\r\x12\x13\n\x0b\x65\x61se_factor\x18\x03 \x01(\x02\x12\x0e\n\x06lapses\x18\x04 \x01(\r\x12\x0f\n\x07leeched\x18\x05 \x01(\x08\x12\x36\n\x0cmemory_state\x18\x06 \x01(\x0b\x32\x1b.anki.cards.FsrsMemoryStateH\x00\x88\x01\x01\x42\x0f\n\r_memory_state\x1a\x80\x01\n\nRelearning\x12\x36\n\x06review\x18\x01 \x01(\x0b\x32&.anki.scheduler.SchedulingState.Review\x12:\n\x08learning\x18\x02 \x01(\x0b\x32(.anki.scheduler.SchedulingState.Learning\x1a\xfe\x01\n\x06Normal\x12\x32\n\x03new\x18\x01 \x01(\x0b\x32#.anki.scheduler.SchedulingState.NewH\x00\x12<\n\x08learning\x18\x02 \x01(\x0b\x32(.anki.scheduler.SchedulingState.LearningH\x00\x12\x38\n\x06review\x18\x03 \x01(\x0b\x32&.anki.scheduler.SchedulingState.ReviewH\x00\x12@\n\nrelearning\x18\x04 \x01(\x0b\x32*.anki.scheduler.SchedulingState.RelearningH\x00\x42\x06\n\x04kind\x1a\x33\n\x07Preview\x12\x16\n\x0escheduled_secs\x18\x01 \x01(\r\x12\x10\n\x08\x66inished\x18\x02 \x01(\x08\x1aT\n\x12ReschedulingFilter\x12>\n\x0eoriginal_state\x18\x01 \x01(\x0b\x32&.anki.scheduler.SchedulingState.Normal\x1a\x9a\x01\n\x08\x46iltered\x12:\n\x07preview\x18\x01 \x01(\x0b\x32\'.anki.scheduler.SchedulingState.PreviewH\x00\x12J\n\x0crescheduling\x18\x02 \x01(\x0b\x32\x32.anki.scheduler.SchedulingState.ReschedulingFilterH\x00\x42\x06\n\x04kindB\x06\n\x04kindB\x0e\n\x0c_custom_data\"\xf8\x02\n\x0bQueuedCards\x12\x35\n\x05\x63\x61rds\x18\x01 \x03(\x0b\x32&.anki.scheduler.QueuedCards.QueuedCard\x12\x11\n\tnew_count\x18\x02 \x01(\r\x12\x16\n\x0elearning_count\x18\x03 \x01(\r\x12\x14\n\x0creview_count\x18\x04 \x01(\r\x1a\xc4\x01\n\nQueuedCard\x12\x1e\n\x04\x63\x61rd\x18\x01 \x01(\x0b\x32\x10.anki.cards.Card\x12\x30\n\x05queue\x18\x02 \x01(\x0e\x32!.anki.scheduler.QueuedCards.Queue\x12\x30\n\x06states\x18\x03 \x01(\x0b\x32 .anki.scheduler.SchedulingStates\x12\x32\n\x07\x63ontext\x18\x04 \x01(\x0b\x32!.anki.scheduler.SchedulingContext\"*\n\x05Queue\x12\x07\n\x03NEW\x10\x00\x12\x0c\n\x08LEARNING\x10\x01\x12\n\n\x06REVIEW\x10\x02\"L\n\x15GetQueuedCardsRequest\x12\x13\n\x0b\x66\x65tch_limit\x18\x01 \x01(\r\x12\x1e\n\x16intraday_learning_only\x18\x02 \x01(\x08\"E\n\x18SchedTimingTodayResponse\x12\x14\n\x0c\x64\x61ys_elapsed\x18\x01 \x01(\r\x12\x13\n\x0bnext_day_at\x18\x02 \x01(\x03\"<\n\x1aStudiedTodayMessageRequest\x12\r\n\x05\x63\x61rds\x18\x01 \x01(\r\x12\x0f\n\x07seconds\x18\x02 \x01(\x01\"i\n\x12UpdateStatsRequest\x12\x0f\n\x07\x64\x65\x63k_id\x18\x01 \x01(\x03\x12\x11\n\tnew_delta\x18\x02 \x01(\x05\x12\x14\n\x0creview_delta\x18\x04 \x01(\x05\x12\x19\n\x11millisecond_delta\x18\x05 \x01(\x05\"O\n\x13\x45xtendLimitsRequest\x12\x0f\n\x07\x64\x65\x63k_id\x18\x01 \x01(\x03\x12\x11\n\tnew_delta\x18\x02 \x01(\x05\x12\x14\n\x0creview_delta\x18\x03 \x01(\x05\"9\n\x1a\x43ountsForDeckTodayResponse\x12\x0b\n\x03new\x18\x01 \x01(\x05\x12\x0e\n\x06review\x18\x02 \x01(\x05\"\x8b\x02\n\x14\x43ongratsInfoResponse\x12\x17\n\x0flearn_remaining\x18\x01 \x01(\r\x12\x1d\n\x15secs_until_next_learn\x18\x02 \x01(\r\x12\x18\n\x10review_remaining\x18\x03 \x01(\x08\x12\x15\n\rnew_remaining\x18\x04 \x01(\x08\x12\x19\n\x11have_sched_buried\x18\x05 \x01(\x08\x12\x18\n\x10have_user_buried\x18\x06 \x01(\x08\x12\x18\n\x10is_filtered_deck\x18\x07 \x01(\x08\x12!\n\x19\x62ridge_commands_supported\x18\x08 \x01(\x08\x12\x18\n\x10\x64\x65\x63k_description\x18\t \x01(\t\"\x8a\x01\n\x11UnburyDeckRequest\x12\x0f\n\x07\x64\x65\x63k_id\x18\x01 \x01(\x03\x12\x34\n\x04mode\x18\x02 \x01(\x0e\x32&.anki.scheduler.UnburyDeckRequest.Mode\".\n\x04Mode\x12\x07\n\x03\x41LL\x10\x00\x12\x0e\n\nSCHED_ONLY\x10\x01\x12\r\n\tUSER_ONLY\x10\x02\"\xb1\x01\n\x19\x42uryOrSuspendCardsRequest\x12\x10\n\x08\x63\x61rd_ids\x18\x01 \x03(\x03\x12\x10\n\x08note_ids\x18\x02 \x03(\x03\x12<\n\x04mode\x18\x03 \x01(\x0e\x32..anki.scheduler.BuryOrSuspendCardsRequest.Mode\"2\n\x04Mode\x12\x0b\n\x07SUSPEND\x10\x00\x12\x0e\n\nBURY_SCHED\x10\x01\x12\r\n\tBURY_USER\x10\x02\"\xe5\x01\n\x19ScheduleCardsAsNewRequest\x12\x10\n\x08\x63\x61rd_ids\x18\x01 \x03(\x03\x12\x0b\n\x03log\x18\x02 \x01(\x08\x12\x18\n\x10restore_position\x18\x03 \x01(\x08\x12\x14\n\x0creset_counts\x18\x04 \x01(\x08\x12G\n\x07\x63ontext\x18\x05 \x01(\x0e\x32\x31.anki.scheduler.ScheduleCardsAsNewRequest.ContextH\x00\x88\x01\x01\"$\n\x07\x43ontext\x12\x0b\n\x07\x42ROWSER\x10\x00\x12\x0c\n\x08REVIEWER\x10\x01\x42\n\n\x08_context\"g\n!ScheduleCardsAsNewDefaultsRequest\x12\x42\n\x07\x63ontext\x18\x01 \x01(\x0e\x32\x31.anki.scheduler.ScheduleCardsAsNewRequest.Context\"T\n\"ScheduleCardsAsNewDefaultsResponse\x12\x18\n\x10restore_position\x18\x01 \x01(\x08\x12\x14\n\x0creset_counts\x18\x02 \x01(\x08\"m\n\x11SetDueDateRequest\x12\x10\n\x08\x63\x61rd_ids\x18\x01 \x03(\x03\x12\x0c\n\x04\x64\x61ys\x18\x02 \x01(\t\x12\x38\n\nconfig_key\x18\x03 \x01(\x0b\x32$.anki.config.OptionalStringConfigKey\"V\n\x0fGradeNowRequest\x12\x10\n\x08\x63\x61rd_ids\x18\x01 \x03(\x03\x12\x31\n\x06rating\x18\x02 \x01(\x0e\x32!.anki.scheduler.CardAnswer.Rating\"y\n\x10SortCardsRequest\x12\x10\n\x08\x63\x61rd_ids\x18\x01 \x03(\x03\x12\x15\n\rstarting_from\x18\x02 \x01(\r\x12\x11\n\tstep_size\x18\x03 \x01(\r\x12\x11\n\trandomize\x18\x04 \x01(\x08\x12\x16\n\x0eshift_existing\x18\x05 \x01(\x08\"5\n\x0fSortDeckRequest\x12\x0f\n\x07\x64\x65\x63k_id\x18\x01 \x01(\x03\x12\x11\n\trandomize\x18\x02 \x01(\x08\"\x81\x02\n\x10SchedulingStates\x12\x30\n\x07\x63urrent\x18\x01 \x01(\x0b\x32\x1f.anki.scheduler.SchedulingState\x12.\n\x05\x61gain\x18\x02 \x01(\x0b\x32\x1f.anki.scheduler.SchedulingState\x12-\n\x04hard\x18\x03 \x01(\x0b\x32\x1f.anki.scheduler.SchedulingState\x12-\n\x04good\x18\x04 \x01(\x0b\x32\x1f.anki.scheduler.SchedulingState\x12-\n\x04\x65\x61sy\x18\x05 \x01(\x0b\x32\x1f.anki.scheduler.SchedulingState\"\xa7\x02\n\nCardAnswer\x12\x0f\n\x07\x63\x61rd_id\x18\x01 \x01(\x03\x12\x36\n\rcurrent_state\x18\x02 \x01(\x0b\x32\x1f.anki.scheduler.SchedulingState\x12\x32\n\tnew_state\x18\x03 \x01(\x0b\x32\x1f.anki.scheduler.SchedulingState\x12\x31\n\x06rating\x18\x04 \x01(\x0e\x32!.anki.scheduler.CardAnswer.Rating\x12\x1a\n\x12\x61nswered_at_millis\x18\x05 \x01(\x03\x12\x1a\n\x12milliseconds_taken\x18\x06 \x01(\r\"1\n\x06Rating\x12\t\n\x05\x41GAIN\x10\x00\x12\x08\n\x04HARD\x10\x01\x12\x08\n\x04GOOD\x10\x02\x12\x08\n\x04\x45\x41SY\x10\x03\"\xd6\x03\n\x12\x43ustomStudyRequest\x12\x0f\n\x07\x64\x65\x63k_id\x18\x01 \x01(\x03\x12\x19\n\x0fnew_limit_delta\x18\x02 \x01(\x05H\x00\x12\x1c\n\x12review_limit_delta\x18\x03 \x01(\x05H\x00\x12\x15\n\x0b\x66orgot_days\x18\x04 \x01(\rH\x00\x12\x1b\n\x11review_ahead_days\x18\x05 \x01(\rH\x00\x12\x16\n\x0cpreview_days\x18\x06 \x01(\rH\x00\x12\x37\n\x04\x63ram\x18\x07 \x01(\x0b\x32\'.anki.scheduler.CustomStudyRequest.CramH\x00\x1a\xe7\x01\n\x04\x43ram\x12>\n\x04kind\x18\x01 \x01(\x0e\x32\x30.anki.scheduler.CustomStudyRequest.Cram.CramKind\x12\x12\n\ncard_limit\x18\x02 \x01(\r\x12\x17\n\x0ftags_to_include\x18\x03 \x03(\t\x12\x17\n\x0ftags_to_exclude\x18\x04 \x03(\t\"Y\n\x08\x43ramKind\x12\x11\n\rCRAM_KIND_DUE\x10\x00\x12\x11\n\rCRAM_KIND_NEW\x10\x01\x12\x14\n\x10\x43RAM_KIND_REVIEW\x10\x02\x12\x11\n\rCRAM_KIND_ALL\x10\x03\x42\x07\n\x05value\"4\n\x11SchedulingContext\x12\x11\n\tdeck_name\x18\x01 \x01(\t\x12\x0c\n\x04seed\x18\x02 \x01(\x04\"-\n\x1a\x43ustomStudyDefaultsRequest\x12\x0f\n\x07\x64\x65\x63k_id\x18\x01 \x01(\x03\"\xb8\x02\n\x1b\x43ustomStudyDefaultsResponse\x12=\n\x04tags\x18\x01 \x03(\x0b\x32/.anki.scheduler.CustomStudyDefaultsResponse.Tag\x12\x12\n\nextend_new\x18\x02 \x01(\r\x12\x15\n\rextend_review\x18\x03 \x01(\r\x12\x15\n\ravailable_new\x18\x04 \x01(\r\x12\x18\n\x10\x61vailable_review\x18\x05 \x01(\r\x12!\n\x19\x61vailable_new_in_children\x18\x06 \x01(\r\x12$\n\x1c\x61vailable_review_in_children\x18\x07 \x01(\r\x1a\x35\n\x03Tag\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x07include\x18\x02 \x01(\x08\x12\x0f\n\x07\x65xclude\x18\x03 \x01(\x08\";\n\x1aRepositionDefaultsResponse\x12\x0e\n\x06random\x18\x01 \x01(\x08\x12\r\n\x05shift\x18\x02 \x01(\x08\"\x9b\x01\n\x18\x43omputeFsrsParamsRequest\x12\x0e\n\x06search\x18\x01 \x01(\t\x12\x16\n\x0e\x63urrent_params\x18\x02 \x03(\x02\x12 \n\x18ignore_revlogs_before_ms\x18\x03 \x01(\x03\x12\x1f\n\x17num_of_relearning_steps\x18\x04 \x01(\r\x12\x14\n\x0chealth_check\x18\x05 \x01(\x08\"y\n\x19\x43omputeFsrsParamsResponse\x12\x0e\n\x06params\x18\x01 \x03(\x02\x12\x12\n\nfsrs_items\x18\x02 \x01(\r\x12 \n\x13health_check_passed\x18\x03 \x01(\x08H\x00\x88\x01\x01\x42\x16\n\x14_health_check_passed\"L\n!ComputeFsrsParamsFromItemsRequest\x12\'\n\x05items\x18\x01 \x03(\x0b\x32\x18.anki.scheduler.FsrsItem\"C\n\x14\x46srsBenchmarkRequest\x12+\n\ttrain_set\x18\x01 \x03(\x0b\x32\x18.anki.scheduler.FsrsItem\"\'\n\x15\x46srsBenchmarkResponse\x12\x0e\n\x06params\x18\x01 \x03(\x02\"@\n\x14\x45xportDatasetRequest\x12\x13\n\x0bmin_entries\x18\x01 \x01(\r\x12\x13\n\x0btarget_path\x18\x02 \x01(\t\"7\n\x08\x46srsItem\x12+\n\x07reviews\x18\x01 \x03(\x0b\x32\x1a.anki.scheduler.FsrsReview\"-\n\nFsrsReview\x12\x0e\n\x06rating\x18\x01 \x01(\r\x12\x0f\n\x07\x64\x65lta_t\x18\x02 \x01(\r\"\xf3\x03\n\x19SimulateFsrsReviewRequest\x12\x0e\n\x06params\x18\x01 \x03(\x02\x12\x19\n\x11\x64\x65sired_retention\x18\x02 \x01(\x02\x12\x11\n\tdeck_size\x18\x03 \x01(\r\x12\x18\n\x10\x64\x61ys_to_simulate\x18\x04 \x01(\r\x12\x11\n\tnew_limit\x18\x05 \x01(\r\x12\x14\n\x0creview_limit\x18\x06 \x01(\r\x12\x14\n\x0cmax_interval\x18\x07 \x01(\r\x12\x0e\n\x06search\x18\x08 \x01(\t\x12%\n\x1dnew_cards_ignore_review_limit\x18\t \x01(\x08\x12\x1d\n\x15\x65\x61sy_days_percentages\x18\n \x03(\x02\x12I\n\x0creview_order\x18\x0b \x01(\x0e\x32\x33.anki.deck_config.DeckConfig.Config.ReviewCardOrder\x12&\n\x19suspend_after_lapse_count\x18\x0c \x01(\rH\x00\x88\x01\x01\x12\x1c\n\x14historical_retention\x18\r \x01(\x02\x12\x1b\n\x13learning_step_count\x18\x0e \x01(\r\x12\x1d\n\x15relearning_step_count\x18\x0f \x01(\rB\x1c\n\x1a_suspend_after_lapse_count\"\x95\x01\n\x1aSimulateFsrsReviewResponse\x12)\n!accumulated_knowledge_acquisition\x18\x01 \x03(\x02\x12\x1a\n\x12\x64\x61ily_review_count\x18\x02 \x03(\r\x12\x17\n\x0f\x64\x61ily_new_count\x18\x03 \x03(\r\x12\x17\n\x0f\x64\x61ily_time_cost\x18\x04 \x03(\x02\"\x9c\x03\n\x1cSimulateFsrsWorkloadResponse\x12\x44\n\x04\x63ost\x18\x01 \x03(\x0b\x32\x36.anki.scheduler.SimulateFsrsWorkloadResponse.CostEntry\x12N\n\tmemorized\x18\x02 \x03(\x0b\x32;.anki.scheduler.SimulateFsrsWorkloadResponse.MemorizedEntry\x12S\n\x0creview_count\x18\x03 \x03(\x0b\x32=.anki.scheduler.SimulateFsrsWorkloadResponse.ReviewCountEntry\x1a+\n\tCostEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\x1a\x30\n\x0eMemorizedEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\x02:\x02\x38\x01\x1a\x32\n\x10ReviewCountEntry\x12\x0b\n\x03key\x18\x01 \x01(\r\x12\r\n\x05value\x18\x02 \x01(\r:\x02\x38\x01\"<\n\x1f\x43omputeOptimalRetentionResponse\x12\x19\n\x11optimal_retention\x18\x01 \x01(\x02\"6\n$GetOptimalRetentionParametersRequest\x12\x0e\n\x06search\x18\x01 \x01(\t\"\x91\x03\n%GetOptimalRetentionParametersResponse\x12\x11\n\tdeck_size\x18\x01 \x01(\r\x12\x12\n\nlearn_span\x18\x02 \x01(\r\x12\x17\n\x0fmax_cost_perday\x18\x03 \x01(\x02\x12\x0f\n\x07max_ivl\x18\x04 \x01(\x02\x12\x19\n\x11\x66irst_rating_prob\x18\x05 \x03(\x02\x12\x1a\n\x12review_rating_prob\x18\x06 \x03(\x02\x12\x15\n\rloss_aversion\x18\x07 \x01(\x02\x12\x13\n\x0blearn_limit\x18\x08 \x01(\r\x12\x14\n\x0creview_limit\x18\t \x01(\r\x12!\n\x19learning_step_transitions\x18\n \x03(\x02\x12#\n\x1brelearning_step_transitions\x18\x0b \x03(\x02\x12\x1a\n\x12state_rating_costs\x18\x0c \x03(\x02\x12\x1b\n\x13learning_step_count\x18\r \x01(\r\x12\x1d\n\x15relearning_step_count\x18\x0e \x01(\r\"j\n\x15\x45valuateParamsRequest\x12\x0e\n\x06search\x18\x01 \x01(\t\x12 \n\x18ignore_revlogs_before_ms\x18\x02 \x01(\x03\x12\x1f\n\x17num_of_relearning_steps\x18\x03 \x01(\r\"_\n\x1b\x45valuateParamsLegacyRequest\x12\x0e\n\x06params\x18\x01 \x03(\x02\x12\x0e\n\x06search\x18\x02 \x01(\t\x12 \n\x18ignore_revlogs_before_ms\x18\x03 \x01(\x03\"=\n\x16\x45valuateParamsResponse\x12\x10\n\x08log_loss\x18\x01 \x01(\x02\x12\x11\n\trmse_bins\x18\x02 \x01(\x02\"\x81\x01\n\x1a\x43omputeMemoryStateResponse\x12/\n\x05state\x18\x01 \x01(\x0b\x32\x1b.anki.cards.FsrsMemoryStateH\x00\x88\x01\x01\x12\x19\n\x11\x64\x65sired_retention\x18\x02 \x01(\x02\x12\r\n\x05\x64\x65\x63\x61y\x18\x03 \x01(\x02\x42\x08\n\x06_state\"5\n\x10\x46uzzDeltaRequest\x12\x0f\n\x07\x63\x61rd_id\x18\x01 \x01(\x03\x12\x10\n\x08interval\x18\x02 \x01(\r\"\'\n\x11\x46uzzDeltaResponse\x12\x12\n\ndelta_days\x18\x01 \x01(\x11\x32\xfe\x18\n\x10SchedulerService\x12T\n\x0eGetQueuedCards\x12%.anki.scheduler.GetQueuedCardsRequest\x1a\x1b.anki.scheduler.QueuedCards\x12\x44\n\nAnswerCard\x12\x1a.anki.scheduler.CardAnswer\x1a\x1a.anki.collection.OpChanges\x12Q\n\x10SchedTimingToday\x12\x13.anki.generic.Empty\x1a(.anki.scheduler.SchedTimingTodayResponse\x12\x39\n\x0cStudiedToday\x12\x13.anki.generic.Empty\x1a\x14.anki.generic.String\x12W\n\x13StudiedTodayMessage\x12*.anki.scheduler.StudiedTodayMessageRequest\x1a\x14.anki.generic.String\x12\x46\n\x0bUpdateStats\x12\".anki.scheduler.UpdateStatsRequest\x1a\x13.anki.generic.Empty\x12H\n\x0c\x45xtendLimits\x12#.anki.scheduler.ExtendLimitsRequest\x1a\x13.anki.generic.Empty\x12T\n\x12\x43ountsForDeckToday\x12\x12.anki.decks.DeckId\x1a*.anki.scheduler.CountsForDeckTodayResponse\x12I\n\x0c\x43ongratsInfo\x12\x13.anki.generic.Empty\x1a$.anki.scheduler.CongratsInfoResponse\x12Q\n\x1eRestoreBuriedAndSuspendedCards\x12\x13.anki.cards.CardIds\x1a\x1a.anki.collection.OpChanges\x12K\n\nUnburyDeck\x12!.anki.scheduler.UnburyDeckRequest\x1a\x1a.anki.collection.OpChanges\x12\x64\n\x12\x42uryOrSuspendCards\x12).anki.scheduler.BuryOrSuspendCardsRequest\x1a#.anki.collection.OpChangesWithCount\x12\x43\n\x11\x45mptyFilteredDeck\x12\x12.anki.decks.DeckId\x1a\x1a.anki.collection.OpChanges\x12N\n\x13RebuildFilteredDeck\x12\x12.anki.decks.DeckId\x1a#.anki.collection.OpChangesWithCount\x12[\n\x12ScheduleCardsAsNew\x12).anki.scheduler.ScheduleCardsAsNewRequest\x1a\x1a.anki.collection.OpChanges\x12\x83\x01\n\x1aScheduleCardsAsNewDefaults\x12\x31.anki.scheduler.ScheduleCardsAsNewDefaultsRequest\x1a\x32.anki.scheduler.ScheduleCardsAsNewDefaultsResponse\x12K\n\nSetDueDate\x12!.anki.scheduler.SetDueDateRequest\x1a\x1a.anki.collection.OpChanges\x12G\n\x08GradeNow\x12\x1f.anki.scheduler.GradeNowRequest\x1a\x1a.anki.collection.OpChanges\x12R\n\tSortCards\x12 .anki.scheduler.SortCardsRequest\x1a#.anki.collection.OpChangesWithCount\x12P\n\x08SortDeck\x12\x1f.anki.scheduler.SortDeckRequest\x1a#.anki.collection.OpChangesWithCount\x12K\n\x13GetSchedulingStates\x12\x12.anki.cards.CardId\x1a .anki.scheduler.SchedulingStates\x12P\n\x12\x44\x65scribeNextStates\x12 .anki.scheduler.SchedulingStates\x1a\x18.anki.generic.StringList\x12\x43\n\x0cStateIsLeech\x12\x1f.anki.scheduler.SchedulingState\x1a\x12.anki.generic.Bool\x12<\n\x10UpgradeScheduler\x12\x13.anki.generic.Empty\x1a\x13.anki.generic.Empty\x12M\n\x0b\x43ustomStudy\x12\".anki.scheduler.CustomStudyRequest\x1a\x1a.anki.collection.OpChanges\x12n\n\x13\x43ustomStudyDefaults\x12*.anki.scheduler.CustomStudyDefaultsRequest\x1a+.anki.scheduler.CustomStudyDefaultsResponse\x12U\n\x12RepositionDefaults\x12\x13.anki.generic.Empty\x1a*.anki.scheduler.RepositionDefaultsResponse\x12h\n\x11\x43omputeFsrsParams\x12(.anki.scheduler.ComputeFsrsParamsRequest\x1a).anki.scheduler.ComputeFsrsParamsResponse\x12\x8c\x01\n\x1dGetOptimalRetentionParameters\x12\x34.anki.scheduler.GetOptimalRetentionParametersRequest\x1a\x35.anki.scheduler.GetOptimalRetentionParametersResponse\x12u\n\x17\x43omputeOptimalRetention\x12).anki.scheduler.SimulateFsrsReviewRequest\x1a/.anki.scheduler.ComputeOptimalRetentionResponse\x12k\n\x12SimulateFsrsReview\x12).anki.scheduler.SimulateFsrsReviewRequest\x1a*.anki.scheduler.SimulateFsrsReviewResponse\x12o\n\x14SimulateFsrsWorkload\x12).anki.scheduler.SimulateFsrsReviewRequest\x1a,.anki.scheduler.SimulateFsrsWorkloadResponse\x12_\n\x0e\x45valuateParams\x12%.anki.scheduler.EvaluateParamsRequest\x1a&.anki.scheduler.EvaluateParamsResponse\x12k\n\x14\x45valuateParamsLegacy\x12+.anki.scheduler.EvaluateParamsLegacyRequest\x1a&.anki.scheduler.EvaluateParamsResponse\x12T\n\x12\x43omputeMemoryState\x12\x12.anki.cards.CardId\x1a*.anki.scheduler.ComputeMemoryStateResponse\x12P\n\tFuzzDelta\x12 .anki.scheduler.FuzzDeltaRequest\x1a!.anki.scheduler.FuzzDeltaResponse2\xbf\x02\n\x17\x42\x61\x63kendSchedulerService\x12z\n\x1a\x43omputeFsrsParamsFromItems\x12\x31.anki.scheduler.ComputeFsrsParamsFromItemsRequest\x1a).anki.scheduler.ComputeFsrsParamsResponse\x12\\\n\rFsrsBenchmark\x12$.anki.scheduler.FsrsBenchmarkRequest\x1a%.anki.scheduler.FsrsBenchmarkResponse\x12J\n\rExportDataset\x12$.anki.scheduler.ExportDatasetRequest\x1a\x13.anki.generic.EmptyB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.scheduler_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_SIMULATEFSRSWORKLOADRESPONSE_COSTENTRY']._loaded_options = None
  _globals['_SIMULATEFSRSWORKLOADRESPONSE_COSTENTRY']._serialized_options = b'8\001'
  _globals['_SIMULATEFSRSWORKLOADRESPONSE_MEMORIZEDENTRY']._loaded_options = None
  _globals['_SIMULATEFSRSWORKLOADRESPONSE_MEMORIZEDENTRY']._serialized_options = b'8\001'
  _globals['_SIMULATEFSRSWORKLOADRESPONSE_REVIEWCOUNTENTRY']._loaded_options = None
  _globals['_SIMULATEFSRSWORKLOADRESPONSE_REVIEWCOUNTENTRY']._serialized_options = b'8\001'
  _globals['_SCHEDULINGSTATE']._serialized_start=163
  _globals['_SCHEDULINGSTATE']._serialized_end=1400
  _globals['_SCHEDULINGSTATE_NEW']._serialized_start=328
  _globals['_SCHEDULINGSTATE_NEW']._serialized_end=351
  _globals['_SCHEDULINGSTATE_LEARNING']._serialized_start=354
  _globals['_SCHEDULINGSTATE_LEARNING']._serialized_end=508
  _globals['_SCHEDULINGSTATE_REVIEW']._serialized_start=511
  _globals['_SCHEDULINGSTATE_REVIEW']._serialized_end=692
  _globals['_SCHEDULINGSTATE_RELEARNING']._serialized_start=695
  _globals['_SCHEDULINGSTATE_RELEARNING']._serialized_end=823
  _globals['_SCHEDULINGSTATE_NORMAL']._serialized_start=826
  _globals['_SCHEDULINGSTATE_NORMAL']._serialized_end=1080
  _globals['_SCHEDULINGSTATE_PREVIEW']._serialized_start=1082
  _globals['_SCHEDULINGSTATE_PREVIEW']._serialized_end=1133
  _globals['_SCHEDULINGSTATE_RESCHEDULINGFILTER']._serialized_start=1135
  _globals['_SCHEDULINGSTATE_RESCHEDULINGFILTER']._serialized_end=1219
  _globals['_SCHEDULINGSTATE_FILTERED']._serialized_start=1222
  _globals['_SCHEDULINGSTATE_FILTERED']._serialized_end=1376
  _globals['_QUEUEDCARDS']._serialized_start=1403
  _globals['_QUEUEDCARDS']._serialized_end=1779
  _globals['_QUEUEDCARDS_QUEUEDCARD']._serialized_start=1539
  _globals['_QUEUEDCARDS_QUEUEDCARD']._serialized_end=1735
  _globals['_QUEUEDCARDS_QUEUE']._serialized_start=1737
  _globals['_QUEUEDCARDS_QUEUE']._serialized_end=1779
  _globals['_GETQUEUEDCARDSREQUEST']._serialized_start=1781
  _globals['_GETQUEUEDCARDSREQUEST']._serialized_end=1857
  _globals['_SCHEDTIMINGTODAYRESPONSE']._serialized_start=1859
  _globals['_SCHEDTIMINGTODAYRESPONSE']._serialized_end=1928
  _globals['_STUDIEDTODAYMESSAGEREQUEST']._serialized_start=1930
  _globals['_STUDIEDTODAYMESSAGEREQUEST']._serialized_end=1990
  _globals['_UPDATESTATSREQUEST']._serialized_start=1992
  _globals['_UPDATESTATSREQUEST']._serialized_end=2097
  _globals['_EXTENDLIMITSREQUEST']._serialized_start=2099
  _globals['_EXTENDLIMITSREQUEST']._serialized_end=2178
  _globals['_COUNTSFORDECKTODAYRESPONSE']._serialized_start=2180
  _globals['_COUNTSFORDECKTODAYRESPONSE']._serialized_end=2237
  _globals['_CONGRATSINFORESPONSE']._serialized_start=2240
  _globals['_CONGRATSINFORESPONSE']._serialized_end=2507
  _globals['_UNBURYDECKREQUEST']._serialized_start=2510
  _globals['_UNBURYDECKREQUEST']._serialized_end=2648
  _globals['_UNBURYDECKREQUEST_MODE']._serialized_start=2602
  _globals['_UNBURYDECKREQUEST_MODE']._serialized_end=2648
  _globals['_BURYORSUSPENDCARDSREQUEST']._serialized_start=2651
  _globals['_BURYORSUSPENDCARDSREQUEST']._serialized_end=2828
  _globals['_BURYORSUSPENDCARDSREQUEST_MODE']._serialized_start=2778
  _globals['_BURYORSUSPENDCARDSREQUEST_MODE']._serialized_end=2828
  _globals['_SCHEDULECARDSASNEWREQUEST']._serialized_start=2831
  _globals['_SCHEDULECARDSASNEWREQUEST']._serialized_end=3060
  _globals['_SCHEDULECARDSASNEWREQUEST_CONTEXT']._serialized_start=3012
  _globals['_SCHEDULECARDSASNEWREQUEST_CONTEXT']._serialized_end=3048
  _globals['_SCHEDULECARDSASNEWDEFAULTSREQUEST']._serialized_start=3062
  _globals['_SCHEDULECARDSASNEWDEFAULTSREQUEST']._serialized_end=3165
  _globals['_SCHEDULECARDSASNEWDEFAULTSRESPONSE']._serialized_start=3167
  _globals['_SCHEDULECARDSASNEWDEFAULTSRESPONSE']._serialized_end=3251
  _globals['_SETDUEDATEREQUEST']._serialized_start=3253
  _globals['_SETDUEDATEREQUEST']._serialized_end=3362
  _globals['_GRADENOWREQUEST']._serialized_start=3364
  _globals['_GRADENOWREQUEST']._serialized_end=3450
  _globals['_SORTCARDSREQUEST']._serialized_start=3452
  _globals['_SORTCARDSREQUEST']._serialized_end=3573
  _globals['_SORTDECKREQUEST']._serialized_start=3575
  _globals['_SORTDECKREQUEST']._serialized_end=3628
  _globals['_SCHEDULINGSTATES']._serialized_start=3631
  _globals['_SCHEDULINGSTATES']._serialized_end=3888
  _globals['_CARDANSWER']._serialized_start=3891
  _globals['_CARDANSWER']._serialized_end=4186
  _globals['_CARDANSWER_RATING']._serialized_start=4137
  _globals['_CARDANSWER_RATING']._serialized_end=4186
  _globals['_CUSTOMSTUDYREQUEST']._serialized_start=4189
  _globals['_CUSTOMSTUDYREQUEST']._serialized_end=4659
  _globals['_CUSTOMSTUDYREQUEST_CRAM']._serialized_start=4419
  _globals['_CUSTOMSTUDYREQUEST_CRAM']._serialized_end=4650
  _globals['_CUSTOMSTUDYREQUEST_CRAM_CRAMKIND']._serialized_start=4561
  _globals['_CUSTOMSTUDYREQUEST_CRAM_CRAMKIND']._serialized_end=4650
  _globals['_SCHEDULINGCONTEXT']._serialized_start=4661
  _globals['_SCHEDULINGCONTEXT']._serialized_end=4713
  _globals['_CUSTOMSTUDYDEFAULTSREQUEST']._serialized_start=4715
  _globals['_CUSTOMSTUDYDEFAULTSREQUEST']._serialized_end=4760
  _globals['_CUSTOMSTUDYDEFAULTSRESPONSE']._serialized_start=4763
  _globals['_CUSTOMSTUDYDEFAULTSRESPONSE']._serialized_end=5075
  _globals['_CUSTOMSTUDYDEFAULTSRESPONSE_TAG']._serialized_start=5022
  _globals['_CUSTOMSTUDYDEFAULTSRESPONSE_TAG']._serialized_end=5075
  _globals['_REPOSITIONDEFAULTSRESPONSE']._serialized_start=5077
  _globals['_REPOSITIONDEFAULTSRESPONSE']._serialized_end=5136
  _globals['_COMPUTEFSRSPARAMSREQUEST']._serialized_start=5139
  _globals['_COMPUTEFSRSPARAMSREQUEST']._serialized_end=5294
  _globals['_COMPUTEFSRSPARAMSRESPONSE']._serialized_start=5296
  _globals['_COMPUTEFSRSPARAMSRESPONSE']._serialized_end=5417
  _globals['_COMPUTEFSRSPARAMSFROMITEMSREQUEST']._serialized_start=5419
  _globals['_COMPUTEFSRSPARAMSFROMITEMSREQUEST']._serialized_end=5495
  _globals['_FSRSBENCHMARKREQUEST']._serialized_start=5497
  _globals['_FSRSBENCHMARKREQUEST']._serialized_end=5564
  _globals['_FSRSBENCHMARKRESPONSE']._serialized_start=5566
  _globals['_FSRSBENCHMARKRESPONSE']._serialized_end=5605
  _globals['_EXPORTDATASETREQUEST']._serialized_start=5607
  _globals['_EXPORTDATASETREQUEST']._serialized_end=5671
  _globals['_FSRSITEM']._serialized_start=5673
  _globals['_FSRSITEM']._serialized_end=5728
  _globals['_FSRSREVIEW']._serialized_start=5730
  _globals['_FSRSREVIEW']._serialized_end=5775
  _globals['_SIMULATEFSRSREVIEWREQUEST']._serialized_start=5778
  _globals['_SIMULATEFSRSREVIEWREQUEST']._serialized_end=6277
  _globals['_SIMULATEFSRSREVIEWRESPONSE']._serialized_start=6280
  _globals['_SIMULATEFSRSREVIEWRESPONSE']._serialized_end=6429
  _globals['_SIMULATEFSRSWORKLOADRESPONSE']._serialized_start=6432
  _globals['_SIMULATEFSRSWORKLOADRESPONSE']._serialized_end=6844
  _globals['_SIMULATEFSRSWORKLOADRESPONSE_COSTENTRY']._serialized_start=6699
  _globals['_SIMULATEFSRSWORKLOADRESPONSE_COSTENTRY']._serialized_end=6742
  _globals['_SIMULATEFSRSWORKLOADRESPONSE_MEMORIZEDENTRY']._serialized_start=6744
  _globals['_SIMULATEFSRSWORKLOADRESPONSE_MEMORIZEDENTRY']._serialized_end=6792
  _globals['_SIMULATEFSRSWORKLOADRESPONSE_REVIEWCOUNTENTRY']._serialized_start=6794
  _globals['_SIMULATEFSRSWORKLOADRESPONSE_REVIEWCOUNTENTRY']._serialized_end=6844
  _globals['_COMPUTEOPTIMALRETENTIONRESPONSE']._serialized_start=6846
  _globals['_COMPUTEOPTIMALRETENTIONRESPONSE']._serialized_end=6906
  _globals['_GETOPTIMALRETENTIONPARAMETERSREQUEST']._serialized_start=6908
  _globals['_GETOPTIMALRETENTIONPARAMETERSREQUEST']._serialized_end=6962
  _globals['_GETOPTIMALRETENTIONPARAMETERSRESPONSE']._serialized_start=6965
  _globals['_GETOPTIMALRETENTIONPARAMETERSRESPONSE']._serialized_end=7366
  _globals['_EVALUATEPARAMSREQUEST']._serialized_start=7368
  _globals['_EVALUATEPARAMSREQUEST']._serialized_end=7474
  _globals['_EVALUATEPARAMSLEGACYREQUEST']._serialized_start=7476
  _globals['_EVALUATEPARAMSLEGACYREQUEST']._serialized_end=7571
  _globals['_EVALUATEPARAMSRESPONSE']._serialized_start=7573
  _globals['_EVALUATEPARAMSRESPONSE']._serialized_end=7634
  _globals['_COMPUTEMEMORYSTATERESPONSE']._serialized_start=7637
  _globals['_COMPUTEMEMORYSTATERESPONSE']._serialized_end=7766
  _globals['_FUZZDELTAREQUEST']._serialized_start=7768
  _globals['_FUZZDELTAREQUEST']._serialized_end=7821
  _globals['_FUZZDELTARESPONSE']._serialized_start=7823
  _globals['_FUZZDELTARESPONSE']._serialized_end=7862
  _globals['_SCHEDULERSERVICE']._serialized_start=7865
  _globals['_SCHEDULERSERVICE']._serialized_end=11063
  _globals['_BACKENDSCHEDULERSERVICE']._serialized_start=11066
  _globals['_BACKENDSCHEDULERSERVICE']._serialized_end=11385
# @@protoc_insertion_point(module_scope)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html


from __future__ import annotations

from typing import Any

from anki.collection import Collection
from anki.utils import max_id

# Base importer
##########################################################################


class Importer:
    needMapper = False
    needDelimiter = False
    dst: Collection | None

    def __init__(self, col: Collection, file: str) -> None:
        self.file = file
        self.log: list[str] = []
        self.col = col.weakref()
        self.total = 0
        self.dst = None

    def run(self) -> None:
        pass

    def open(self) -> None:
        "Open file and ensure it's in the right format."
        return

    def close(self) -> None:
        "Closes the open file."
        return

    # Timestamps
    ######################################################################
    # It's too inefficient to check for existing ids on every object,
    # and a previous import may have created timestamps in the future, so we
    # need to make sure our starting point is safe.

    def _prepareTS(self) -> None:
        self._ts = max_id(self.dst.db)

    def ts(self) -> Any:
        self._ts += 1
        return self._ts



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html


from __future__ import annotations

import json
import os
import unicodedata
import zipfile
from typing import Any

from anki.importing.anki2 import Anki2Importer, MediaMapInvalid
from anki.utils import tmpfile


class AnkiPackageImporter(Anki2Importer):
    nameToNum: dict[str, str]
    zip: zipfile.ZipFile | None

    def run(self) -> None:  # type: ignore
        # extract the deck from the zip file
        self.zip = z = zipfile.ZipFile(self.file)
        # v2 scheduler?
        try:
            z.getinfo("collection.anki21")
            suffix = ".anki21"
        except KeyError:
            suffix = ".anki2"

        col = z.read(f"collection{suffix}")
        colpath = tmpfile(suffix=".anki2")
        with open(colpath, "wb") as f:
            f.write(col)
        self.file = colpath
        # we need the media dict in advance, and we'll need a map of fname ->
        # number to use during the import
        self.nameToNum = {}
        dir = self.col.media.dir()
        try:
            media_dict = json.loads(z.read("media").decode("utf8"))
        except Exception as exc:
            raise MediaMapInvalid() from exc
        for k, v in list(media_dict.items()):
            path = os.path.abspath(os.path.join(dir, v))
            if os.path.commonprefix([path, dir]) != dir:
                raise Exception("Invalid file")

            self.nameToNum[unicodedata.normalize("NFC", v)] = k
        # run anki2 importer
        Anki2Importer.run(self, importing_v2=suffix == ".anki21")
        # import static media
        for file, c in list(self.nameToNum.items()):
            if not file.startswith("_") and not file.startswith("latex-"):
                continue
            path = os.path.join(self.col.media.dir(), file)
            if not os.path.exists(path):
                with open(path, "wb") as f:
                    f.write(z.read(c))

    def _srcMediaData(self, fname: str) -> Any:
        if fname in self.nameToNum:
            return self.zip.read(
                self.nameToNum[fname]
            )  # pytype: disable=attribute-error
        return None



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: anki/config.proto
# Protobuf Python Version: 7.34.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    7,
    34,
    1,
    '',
    'anki/config.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from anki import generic_pb2 as anki_dot_generic__pb2
from anki import collection_pb2 as anki_dot_collection__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x11\x61nki/config.proto\x12\x0b\x61nki.config\x1a\x12\x61nki/generic.proto\x1a\x15\x61nki/collection.proto\"\xdc\x06\n\tConfigKey\"\xe5\x05\n\x04\x42ool\x12!\n\x1d\x42ROWSER_TABLE_SHOW_NOTES_MODE\x10\x00\x12\x16\n\x12PREVIEW_BOTH_SIDES\x10\x03\x12\x11\n\rCOLLAPSE_TAGS\x10\x04\x12\x16\n\x12\x43OLLAPSE_NOTETYPES\x10\x05\x12\x12\n\x0e\x43OLLAPSE_DECKS\x10\x06\x12\x1b\n\x17\x43OLLAPSE_SAVED_SEARCHES\x10\x07\x12\x12\n\x0e\x43OLLAPSE_TODAY\x10\x08\x12\x17\n\x13\x43OLLAPSE_CARD_STATE\x10\t\x12\x12\n\x0e\x43OLLAPSE_FLAGS\x10\n\x12\x0e\n\nSCHED_2021\x10\x0b\x12#\n\x1f\x41\x44\x44ING_DEFAULTS_TO_CURRENT_DECK\x10\x0c\x12\x1b\n\x17HIDE_AUDIO_PLAY_BUTTONS\x10\r\x12\"\n\x1eINTERRUPT_AUDIO_WHEN_ANSWERING\x10\x0e\x12\x17\n\x13PASTE_IMAGES_AS_PNG\x10\x0f\x12\x1b\n\x17PASTE_STRIPS_FORMATTING\x10\x10\x12\x17\n\x13NORMALIZE_NOTE_TEXT\x10\x11\x12\x1c\n\x18IGNORE_ACCENTS_IN_SEARCH\x10\x12\x12\x1c\n\x18RESTORE_POSITION_BROWSER\x10\x13\x12\x1d\n\x19RESTORE_POSITION_REVIEWER\x10\x14\x12\x18\n\x14RESET_COUNTS_BROWSER\x10\x15\x12\x19\n\x15RESET_COUNTS_REVIEWER\x10\x16\x12\x1b\n\x17RANDOM_ORDER_REPOSITION\x10\x17\x12$\n SHIFT_POSITION_OF_EXISTING_CARDS\x10\x18\x12\x10\n\x0cRENDER_LATEX\x10\x19\x12\x19\n\x15LOAD_BALANCER_ENABLED\x10\x1a\x12&\n\"FSRS_SHORT_TERM_WITH_STEPS_ENABLED\x10\x1b\x12\x18\n\x14\x46SRS_LEGACY_EVALUATE\x10\x1c\"g\n\x06String\x12\x13\n\x0fSET_DUE_BROWSER\x10\x00\x12\x14\n\x10SET_DUE_REVIEWER\x10\x01\x12\x17\n\x13\x44\x45\x46\x41ULT_SEARCH_TEXT\x10\x02\x12\x19\n\x15\x43\x41RD_STATE_CUSTOMIZER\x10\x03\"@\n\x14GetConfigBoolRequest\x12(\n\x03key\x18\x01 \x01(\x0e\x32\x1b.anki.config.ConfigKey.Bool\"a\n\x14SetConfigBoolRequest\x12(\n\x03key\x18\x01 \x01(\x0e\x32\x1b.anki.config.ConfigKey.Bool\x12\r\n\x05value\x18\x02 \x01(\x08\x12\x10\n\x08undoable\x18\x03 \x01(\x08\"D\n\x16GetConfigStringRequest\x12*\n\x03key\x18\x01 \x01(\x0e\x32\x1d.anki.config.ConfigKey.String\"e\n\x16SetConfigStringRequest\x12*\n\x03key\x18\x01 \x01(\x0e\x32\x1d.anki.config.ConfigKey.String\x12\r\n\x05value\x18\x02 \x01(\t\x12\x10\n\x08undoable\x18\x03 \x01(\x08\"E\n\x17OptionalStringConfigKey\x12*\n\x03key\x18\x01 \x01(\x0e\x32\x1d.anki.config.ConfigKey.String\"I\n\x14SetConfigJsonRequest\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x12\n\nvalue_json\x18\x02 \x01(\x0c\x12\x10\n\x08undoable\x18\x03 \x01(\x08\"\x86\x08\n\x0bPreferences\x12\x37\n\nscheduling\x18\x01 \x01(\x0b\x32#.anki.config.Preferences.Scheduling\x12\x35\n\treviewing\x18\x02 \x01(\x0b\x32\".anki.config.Preferences.Reviewing\x12\x31\n\x07\x65\x64iting\x18\x03 \x01(\x0b\x32 .anki.config.Preferences.Editing\x12\x36\n\x07\x62\x61\x63kups\x18\x04 \x01(\x0b\x32%.anki.config.Preferences.BackupLimits\x1a\xf3\x01\n\nScheduling\x12\x10\n\x08rollover\x18\x02 \x01(\r\x12\x18\n\x10learn_ahead_secs\x18\x03 \x01(\r\x12H\n\x0enew_review_mix\x18\x04 \x01(\x0e\x32\x30.anki.config.Preferences.Scheduling.NewReviewMix\x12\x14\n\x0cnew_timezone\x18\x05 \x01(\x08\x12\x17\n\x0f\x64\x61y_learn_first\x18\x06 \x01(\x08\"@\n\x0cNewReviewMix\x12\x0e\n\nDISTRIBUTE\x10\x00\x12\x11\n\rREVIEWS_FIRST\x10\x01\x12\r\n\tNEW_FIRST\x10\x02\x1a\xfe\x01\n\tReviewing\x12\x1f\n\x17hide_audio_play_buttons\x18\x01 \x01(\x08\x12&\n\x1einterrupt_audio_when_answering\x18\x02 \x01(\x08\x12!\n\x19show_remaining_due_counts\x18\x03 \x01(\x08\x12!\n\x19show_intervals_on_buttons\x18\x04 \x01(\x08\x12\x17\n\x0ftime_limit_secs\x18\x05 \x01(\r\x12\x1d\n\x15load_balancer_enabled\x18\x06 \x01(\x08\x12*\n\"fsrs_short_term_with_steps_enabled\x18\x07 \x01(\x08\x1a\xc5\x01\n\x07\x45\x64iting\x12\'\n\x1f\x61\x64\x64ing_defaults_to_current_deck\x18\x01 \x01(\x08\x12\x1b\n\x13paste_images_as_png\x18\x02 \x01(\x08\x12\x1f\n\x17paste_strips_formatting\x18\x03 \x01(\x08\x12\x1b\n\x13\x64\x65\x66\x61ult_search_text\x18\x04 \x01(\t\x12 \n\x18ignore_accents_in_search\x18\x05 \x01(\x08\x12\x14\n\x0crender_latex\x18\x06 \x01(\x08\x1a]\n\x0c\x42\x61\x63kupLimits\x12\r\n\x05\x64\x61ily\x18\x01 \x01(\r\x12\x0e\n\x06weekly\x18\x02 \x01(\r\x12\x0f\n\x07monthly\x18\x03 \x01(\r\x12\x1d\n\x15minimum_interval_mins\x18\x04 \x01(\r2\xa7\x06\n\rConfigService\x12\x39\n\rGetConfigJson\x12\x14.anki.generic.String\x1a\x12.anki.generic.Json\x12N\n\rSetConfigJson\x12!.anki.config.SetConfigJsonRequest\x1a\x1a.anki.collection.OpChanges\x12M\n\x13SetConfigJsonNoUndo\x12!.anki.config.SetConfigJsonRequest\x1a\x13.anki.generic.Empty\x12@\n\x0cRemoveConfig\x12\x14.anki.generic.String\x1a\x1a.anki.collection.OpChanges\x12\x37\n\x0cGetAllConfig\x12\x13.anki.generic.Empty\x1a\x12.anki.generic.Json\x12\x46\n\rGetConfigBool\x12!.anki.config.GetConfigBoolRequest\x1a\x12.anki.generic.Bool\x12N\n\rSetConfigBool\x12!.anki.config.SetConfigBoolRequest\x1a\x1a.anki.collection.OpChanges\x12L\n\x0fGetConfigString\x12#.anki.config.GetConfigStringRequest\x1a\x14.anki.generic.String\x12R\n\x0fSetConfigString\x12#.anki.config.SetConfigStringRequest\x1a\x1a.anki.collection.OpChanges\x12?\n\x0eGetPreferences\x12\x13.anki.generic.Empty\x1a\x18.anki.config.Preferences\x12\x46\n\x0eSetPreferences\x12\x18.anki.config.Preferences\x1a\x1a.anki.collection.OpChanges2\x16\n\x14\x42\x61\x63kendConfigServiceB\x02P\x01\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'anki.config_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  _globals['DESCRIPTOR']._loaded_options = None
  _globals['DESCRIPTOR']._serialized_options = b'P\001'
  _globals['_CONFIGKEY']._serialized_start=78
  _globals['_CONFIGKEY']._serialized_end=938
  _globals['_CONFIGKEY_BOOL']._serialized_start=92
  _globals['_CONFIGKEY_BOOL']._serialized_end=833
  _globals['_CONFIGKEY_STRING']._serialized_start=835
  _globals['_CONFIGKEY_STRING']._serialized_end=938
  _globals['_GETCONFIGBOOLREQUEST']._serialized_start=940
  _globals['_GETCONFIGBOOLREQUEST']._serialized_end=1004
  _globals['_SETCONFIGBOOLREQUEST']._serialized_start=1006
  _globals['_SETCONFIGBOOLREQUEST']._serialized_end=1103
  _globals['_GETCONFIGSTRINGREQUEST']._serialized_start=1105
  _globals['_GETCONFIGSTRINGREQUEST']._serialized_end=1173
  _globals['_SETCONFIGSTRINGREQUEST']._serialized_start=1175
  _globals['_SETCONFIGSTRINGREQUEST']._serialized_end=1276
  _globals['_OPTIONALSTRINGCONFIGKEY']._serialized_start=1278
  _globals['_OPTIONALSTRINGCONFIGKEY']._serialized_end=1347
  _globals['_SETCONFIGJSONREQUEST']._serialized_start=1349
  _globals['_SETCONFIGJSONREQUEST']._serialized_end=1422
  _globals['_PREFERENCES']._serialized_start=1425
  _globals['_PREFERENCES']._serialized_end=2455
  _globals['_PREFERENCES_SCHEDULING']._serialized_start=1660
  _globals['_PREFERENCES_SCHEDULING']._serialized_end=1903
  _globals['_PREFERENCES_SCHEDULING_NEWREVIEWMIX']._serialized_start=1839
  _globals['_PREFERENCES_SCHEDULING_NEWREVIEWMIX']._serialized_end=1903
  _globals['_PREFERENCES_REVIEWING']._serialized_start=1906
  _globals['_PREFERENCES_REVIEWING']._serialized_end=2160
  _globals['_PREFERENCES_EDITING']._serialized_start=2163
  _globals['_PREFERENCES_EDITING']._serialized_end=2360
  _globals['_PREFERENCES_BACKUPLIMITS']._serialized_start=2362
  _globals['_PREFERENCES_BACKUPLIMITS']._serialized_end=2455
  _globals['_CONFIGSERVICE']._serialized_start=2458
  _globals['_CONFIGSERVICE']._serialized_end=3265
  _globals['_BACKENDCONFIGSERVICE']._serialized_start=3267
  _globals['_BACKENDCONFIGSERVICE']._serialized_end=3289
# @@protoc_insertion_point(module_scope)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import os
import pprint
import re
import sys
import time
from collections.abc import Callable, Sequence

from anki import media_pb2
from anki._legacy import DeprecatedNamesMixin, deprecated_keywords
from anki.consts import *
from anki.latex import render_latex, render_latex_returning_errors
from anki.models import NotetypeId
from anki.sound import SoundOrVideoTag
from anki.template import av_tags_to_native
from anki.utils import int_time


def media_paths_from_col_path(col_path: str) -> tuple[str, str]:
    media_folder = re.sub(r"(?i)\.(anki2)$", ".media", col_path)
    media_db = f"{media_folder}.db2"
    return (media_folder, media_db)


CheckMediaResponse = media_pb2.CheckMediaResponse


class MediaManager(DeprecatedNamesMixin):
    sound_regexps = [r"(?i)(\[sound:(?P<fname>[^]]+)\])"]
    html_media_regexps = [
        # src element quoted case
        r"(?i)(<(?:img|audio|source)\b[^>]* src=(?P<str>[\"'])(?P<fname>[^>]+?)(?P=str)[^>]*>)",
        # unquoted case
        r"(?i)(<(?:img|audio|source)\b[^>]* src=(?!['\"])(?P<fname>[^ >]+)[^>]*?>)",
        # src element quoted case
        r"(?i)(<object\b[^>]* data=(?P<str>[\"'])(?P<fname>[^>]+?)(?P=str)[^>]*>)",
        # unquoted case
        r"(?i)(<object\b[^>]* data=(?!['\"])(?P<fname>[^ >]+)[^>]*?>)",
    ]
    regexps = sound_regexps + html_media_regexps

    def __init__(self, col: anki.collection.Collection, server: bool) -> None:
        self.col = col.weakref()
        if server:
            return
        # media directory
        self._dir = media_paths_from_col_path(self.col.path)[0]
        if not os.path.exists(self._dir):
            os.makedirs(self._dir)

    def __repr__(self) -> str:
        dict_ = dict(self.__dict__)
        del dict_["col"]
        return f"{super().__repr__()} {pprint.pformat(dict_, width=300)}"

    def dir(self) -> str:
        return self._dir

    def force_resync(self) -> None:
        try:
            os.unlink(media_paths_from_col_path(self.col.path)[1])
        except FileNotFoundError:
            pass

    def empty_trash(self) -> None:
        self.col._backend.empty_trash()

    def restore_trash(self) -> None:
        self.col._backend.restore_trash()

    def strip_av_tags(self, text: str) -> str:
        return self.col._backend.strip_av_tags(text)

    def _extract_filenames(self, text: str) -> list[str]:
        "This only exists to support a legacy function; do not use."
        out = self.col._backend.extract_av_tags(text=text, question_side=True)
        return [
            x.filename
            for x in av_tags_to_native(out.av_tags)
            if isinstance(x, SoundOrVideoTag)
        ]

    # File manipulation
    ##########################################################################

    def add_file(self, path: str) -> str:
        """Add basename of path to the media folder, renaming if not unique.

        Returns possibly-renamed filename."""
        with open(path, "rb") as file:
            return self.write_data(os.path.basename(path), file.read())

    def write_data(self, desired_fname: str, data: bytes) -> str:
        """Write the file to the media folder, renaming if not unique.

        Returns possibly-renamed filename."""
        return self.col._backend.add_media_file(desired_name=desired_fname, data=data)

    def add_extension_based_on_mime(self, fname: str, content_type: str) -> str:
        "Add extension based on mime for common audio and image format if missing extension."
        if not os.path.splitext(fname)[1]:
            # mimetypes is returning '.jpe' even after calling .init(), so we'll do
            # it manually instead
            type_map = {
                "audio/mpeg": ".mp3",
                "audio/ogg": ".oga",
                "audio/opus": ".opus",
                "audio/wav": ".wav",
                "audio/webm": ".weba",
                "audio/aac": ".aac",
                "image/jpeg": ".jpg",
                "image/png": ".png",
                "image/svg+xml": ".svg",
                "image/webp": ".webp",
                "image/avif": ".avif",
            }
            if content_type in type_map:
                fname += type_map[content_type]
        return fname

    def have(self, fname: str) -> bool:
        return os.path.exists(os.path.join(self.dir(), fname))

    def trash_files(self, fnames: list[str]) -> None:
        "Move provided files to the trash."
        self.col._backend.trash_media_files(fnames)

    # String manipulation
    ##########################################################################

    @deprecated_keywords(includeRemote="include_remote")
    def files_in_str(
        self, mid: NotetypeId, string: str, include_remote: bool = False
    ) -> list[str]:
        files = []
        model = self.col.models.get(mid)
        # handle latex
        string = render_latex(string, model, self.col)
        # extract filenames
        for reg in self.regexps:
            for match in re.finditer(reg, string):
                fname = match.group("fname")
                is_local = not re.match("(https?|ftp)://", fname.lower())
                if is_local or include_remote:
                    files.append(fname)
        return files

    def extract_static_media_files(self, mid: NotetypeId) -> Sequence[str]:
        return self.col._backend.extract_static_media_files(mid)

    def transform_names(self, txt: str, func: Callable) -> str:
        for reg in self.regexps:
            txt = re.sub(reg, func, txt)
        return txt

    def strip(self, txt: str) -> str:
        "Return text with sound and image tags removed."
        for reg in self.regexps:
            txt = re.sub(reg, "", txt)
        return txt

    def escape_images(self, string: str, unescape: bool = False) -> str:
        "escape_media_filenames alias for compatibility with add-ons."
        return self.escape_media_filenames(string, unescape)

    def escape_media_filenames(self, string: str, unescape: bool = False) -> str:
        "Apply or remove percent encoding to filenames in html tags (audio, image, object)."
        if unescape:
            return self.col._backend.decode_iri_paths(string)
        else:
            return self.col._backend.encode_iri_paths(string)

    # Checking media
    ##########################################################################

    def check(self) -> CheckMediaResponse:
        output = self.col._backend.check_media()
        return output

    def render_all_latex(
        self, progress_cb: Callable[[int], bool] | None = None
    ) -> tuple[int, str] | None:
        """Render any LaTeX that is missing.

        If a progress callback is provided and it returns false, the operation
        will be aborted.

        If an error is encountered, returns (note_id, error_message)
        """
        last_progress = time.time()
        checked = 0
        for nid, mid, flds in self.col.db.execute(
            "select id, mid, flds from notes where flds like '%[%'"
        ):
            model = self.col.models.get(mid)
            _html, errors = render_latex_returning_errors(
                flds, model, self.col, expand_clozes=True
            )
            if errors:
                return (nid, "\n".join(errors))

            checked += 1
            elap = time.time() - last_progress
            if elap >= 0.3 and progress_cb is not None:
                last_progress = int_time()
                if not progress_cb(checked):
                    return None

        return None

    # Legacy
    ##########################################################################

    _illegalCharReg = re.compile(r'[][><:"/?*^\\|\0\r\n]')

    def _legacy_strip_illegal(self, str: str) -> str:
        # currently used by ankiconnect
        return re.sub(self._illegalCharReg, "", str)

    def _legacy_has_illegal(self, string: str) -> bool:
        if re.search(self._illegalCharReg, string):
            return True
        try:
            string.encode(sys.getfilesystemencoding())
        except UnicodeEncodeError:
            return True
        return False

    def _legacy_find_changes(self) -> None:
        pass

    @deprecated_keywords(typeHint="type_hint")
    def _legacy_write_data(
        self, opath: str, data: bytes, type_hint: str | None = None
    ) -> str:
        fname = os.path.basename(opath)
        if type_hint:
            fname = self.add_extension_based_on_mime(fname, type_hint)
        return self.write_data(fname, data)


MediaManager.register_deprecated_attributes(
    stripIllegal=(MediaManager._legacy_strip_illegal, None),
    hasIllegal=(MediaManager._legacy_has_illegal, None),
    findChanges=(MediaManager._legacy_find_changes, None),
    writeData=(MediaManager._legacy_write_data, MediaManager.write_data),
)



# Copyright: Ankitects Pty Ltd and contributors
# License: GNU AGPL, version 3 or later; http://www.gnu.org/licenses/agpl.html

from __future__ import annotations

import aqt
import aqt.forms
import aqt.operations
from anki.collection import OpChangesWithId, SearchNode
from anki.decks import DeckDict, DeckId, FilteredDeckConfig
from anki.errors import SearchError
from anki.lang import without_unicode_isolation
from anki.scheduler import FilteredDeckForUpdate
from aqt import AnkiQt, colors, gui_hooks
from aqt.operations import QueryOp
from aqt.operations.scheduling import add_or_update_filtered_deck
from aqt.qt import *
from aqt.theme import theme_manager
from aqt.utils import (
    HelpPage,
    disable_help_button,
    openHelp,
    restoreGeom,
    saveGeom,
    showWarning,
    tr,
)


class FilteredDeckConfigDialog(QDialog):
    """Dialog to modify and (re)build a filtered deck."""

    GEOMETRY_KEY = "dyndeckconf"
    DIALOG_KEY = "FilteredDeckConfigDialog"
    silentlyClose = True

    def __init__(
        self,
        mw: AnkiQt,
        deck_id: DeckId = DeckId(0),
        search: str | None = None,
        search_2: str | None = None,
    ) -> None:
        """If 'deck_id' is non-zero, load and modify its settings.
        Otherwise, build a new deck and derive settings from the current deck.

        If search or search_2 are provided, they will be used as the default
        search text.
        """

        QDialog.__init__(self, mw)
        self.mw = mw
        mw.garbage_collect_on_dialog_finish(self)
        self.col = self.mw.col
        self._desired_search_1 = search
        self._desired_search_2 = search_2

        self._initial_dialog_setup()

        # set on successful query
        self.deck: FilteredDeckForUpdate

        QueryOp(
            parent=self.mw,
            op=lambda col: col.sched.get_or_create_filtered_deck(deck_id=deck_id),
            success=self.load_deck_and_show,
        ).failure(self.on_fetch_error).run_in_background()

    def on_fetch_error(self, exc: Exception) -> None:
        showWarning(str(exc))
        self.close()

    def _initial_dialog_setup(self) -> None:
        self.form = aqt.forms.filtered_deck.Ui_Dialog()
        self.form.setupUi(self)

        order_labels = self.col.sched.filtered_deck_order_labels()

        self.form.order.addItems(order_labels)
        self.form.order_2.addItems(order_labels)

        qconnect(self.form.allow_empty.stateChanged, self._on_allow_empty_toggled)

        qconnect(self.form.resched.stateChanged, self._onReschedToggled)

        qconnect(self.form.search_button.clicked, self.on_search_button)
        qconnect(self.form.search_button_2.clicked, self.on_search_button_2)
        qconnect(self.form.hint_button.clicked, self.on_hint_button)
        blue = theme_manager.var(colors.FG_LINK)
        grey = theme_manager.var(colors.FG_DISABLED)
        self.setStyleSheet(
            f"""QPushButton[label] {{ padding: 0; border: 0 }}
            QPushButton[label]:hover {{ text-decoration: underline }}
            QPushButton[label="search"] {{ color: {blue} }}
            QPushButton[label="hint"] {{ color: {grey} }}"""
        )
        disable_help_button(self)
        self.setWindowModality(Qt.WindowModality.WindowModal)
        qconnect(
            self.form.buttonBox.helpRequested, lambda: openHelp(HelpPage.FILTERED_DECK)
        )

        self.form.again_delay_label.setText(
            tr.decks_delay_for_button(button=tr.studying_again())
        )
        self.form.hard_delay_label.setText(
            tr.decks_delay_for_button(button=tr.studying_hard())
        )
        self.form.good_delay_label.setText(
            tr.decks_delay_for_button(button=tr.studying_good())
        )

        restoreGeom(self, self.GEOMETRY_KEY)

    def load_deck_and_show(self, deck: FilteredDeckForUpdate) -> None:
        self.deck = deck
        self._load_deck()
        self.show()

    def _load_deck(self) -> None:
        form = self.form
        deck = self.deck
        config = deck.config

        self.form.name.setText(deck.name)
        self.form.name.setPlaceholderText(deck.name)

        existing = deck.id != 0
        if existing:
            build_label = tr.actions_rebuild()
        else:
            build_label = tr.decks_build()

        ok_button = self.form.buttonBox.button(QDialogButtonBox.StandardButton.Ok)
        assert ok_button is not None
        ok_button.setText(build_label)

        form.resched.setChecked(config.reschedule)
        self._onReschedToggled(0)

        term1: FilteredDeckConfig.SearchTerm = config.search_terms[0]
        form.search.setText(term1.search)
        form.order.setCurrentIndex(term1.order)
        form.limit.setValue(term1.limit)

        form.preview_again.setValue(config.preview_again_secs)
        form.preview_hard.setValue(config.preview_hard_secs)
        form.preview_good.setValue(config.preview_good_secs)

        if len(config.search_terms) > 1:
            term2: FilteredDeckConfig.SearchTerm = config.search_terms[1]
            form.search_2.setText(term2.search)
            form.order_2.setCurrentIndex(term2.order)
            form.limit_2.setValue(term2.limit)
            show_second = existing
        else:
            show_second = False
            form.order_2.setCurrentIndex(5)
            form.limit_2.setValue(20)

        form.secondFilter.setChecked(show_second)
        form.filter2group.setVisible(show_second)

        self.set_custom_searches(self._desired_search_1, self._desired_search_2)

        self.setWindowTitle(
            without_unicode_isolation(tr.actions_options_for(val=self.deck.name))
        )

        gui_hooks.filtered_deck_dialog_did_load_deck(self, deck)

    def reopen(
        self,
        _mw: AnkiQt,
        search: str | None = None,
        search_2: str | None = None,
        _deck: DeckDict | None = None,
    ) -> None:
        self.set_custom_searches(search, search_2)

    def set_custom_searches(self, search: str | None, search_2: str | None) -> None:
        if search is not None:
            self.form.search.setText(search)
        self.form.search.setFocus()
        self.form.search.selectAll()
        if search_2 is not None:
            self.form.secondFilter.setChecked(True)
            self.form.filter2group.setVisible(True)
            self.form.search_2.setText(search_2)
            self.form.search_2.setFocus()
            self.form.search_2.selectAll()

    def on_search_button(self) -> None:
        self._on_search_button(self.form.search)

    def on_search_button_2(self) -> None:
        self._on_search_button(self.form.search_2)

    def _on_search_button(self, line: QLineEdit) -> None:
        try:
            search = self.col.build_search_string(line.text())
        except SearchError as err:
            line.setFocus()
            line.selectAll()
            showWarning(str(err))
        else:
            aqt.dialogs.open("Browser", self.mw, search=(search,))

    def on_hint_button(self) -> None:
        """Open the browser to show cards that match the typed-in filters but cannot be included
        due to internal limitations.
        """
        manual_filters = (self.form.search.text(), *self._second_filter())
        implicit_filters = (
            SearchNode(card_state=SearchNode.CARD_STATE_SUSPENDED),
            SearchNode(card_state=SearchNode.CARD_STATE_BURIED),
            *self._filtered_search_node(),
        )
        manual_filter = self.col.group_searches(*manual_filters, joiner="OR")
        implicit_filter = self.col.group_searches(*implicit_filters, joiner="OR")
        try:
            search = self.col.build_search_string(manual_filter, implicit_filter)
        except Exception as err:
            showWarning(str(err))
        else:
            aqt.dialogs.open("Browser", self.mw, search=(search,))

    def _second_filter(self) -> tuple[str, ...]:
        if self.form.secondFilter.isChecked():
            return (self.form.search_2.text(),)
        return ()

    def _filtered_search_node(self) -> tuple[SearchNode]:
        """Return a search node that matches cards in filtered decks, if applicable excluding those
        in the deck being rebuild."""
        if self.deck.id:
            return (
                self.col.group_searches(
                    SearchNode(deck="filtered"),
                    SearchNode(negated=SearchNode(deck=self.deck.name)),
                ),
            )
        return (SearchNode(deck="filtered"),)

    def _onReschedToggled(self, _state: int) -> None:
        self.form.previewDelayWidget.setVisible(not self.form.resched.isChecked())

    def _on_allow_empty_toggled(self) -> None:
        self.deck.allow_empty = self.form.allow_empty.isChecked()

    def _update_deck(self) -> bool:
        """Update our stored deck with the details from the GUI.
        If false, abort adding."""
        form = self.form
        deck = self.deck
        config = deck.config

        deck.name = form.name.text()
        config.reschedule = form.resched.isChecked()

        del config.delays[:]
        terms = [
            FilteredDeckConfig.SearchTerm(
                search=form.search.text(),
                limit=form.limit.value(),
                order=form.order.currentIndex(),  # type: ignore[arg-type]
            )
        ]

        if form.secondFilter.isChecked():
            terms.append(
                FilteredDeckConfig.SearchTerm(
                    search=form.search_2.text(),
                    limit=form.limit_2.value(),
                    order=form.order_2.currentIndex(),  # type: ignore[arg-type]
                )
            )

        del config.search_terms[:]
        config.search_terms.extend(terms)
        config.preview_again_secs = form.preview_again.value()
        config.preview_hard_secs = form.preview_hard.value()
        config.preview_good_secs = form.preview_good.value()

        return True

    def reject(self) -> None:
        aqt.dialogs.markClosed(self.DIALOG_KEY)
        QDialog.reject(self)

    def accept(self) -> None:
        if not self._update_deck():
            return

        def success(out: OpChangesWithId) -> None:
            gui_hooks.filtered_deck_dialog_did_add_or_update_deck(
                self, self.deck, out.id
            )
            saveGeom(self, self.GEOMETRY_KEY)
            aqt.dialogs.markClosed(self.DIALOG_KEY)
            QDialog.accept(self)

        gui_hooks.filtered_deck_dialog_will_add_or_update_deck(self, self.deck)

        add_or_update_filtered_deck(parent=self, deck=self.deck).success(
            success
        ).run_in_background()
